"""
The tunnel community.

Author(s): Egbert Bouman
"""
import os
import random
from asyncio import iscoroutine, sleep
from binascii import unhexlify
from collections import defaultdict

from .caches import *
from .endpoint import TunnelEndpoint
from .payload import *
from .tunnel import *
from .tunnelcrypto import CryptoException, TunnelCrypto
from ...community import Community
from ...lazy_community import lazy_wrapper
from ...messaging.payload_headers import BinMemberAuthenticationPayload
from ...peer import Peer
from ...requestcache import RequestCache
from ...taskmanager import task


def unpack_cell(payload_cls):
    """
    This function wrapper will unpack the normal payload for you, and handle a singular circuit_id parameter at
    the end of the parameter list
    You can now write your non-authenticated and signed functions as follows:
    ::
        @unpack_cell(DataPayload)
        def on_message(source_address, payload):
            '''
            :type source_address: str
            :type payload: DataPayload
            '''
            pass
    """
    def decorator(func):
        def wrapper(self, source_address, data, circuit_id=None):
            payload, _ = self.serializer.unpack_serializable(payload_cls, data, offset=23)
            return func(self, source_address, payload, circuit_id)
        return wrapper
    return decorator


class TunnelSettings(object):

    def __init__(self):
        self.crypto = TunnelCrypto()

        self.min_circuits = 1
        self.max_circuits = 8
        self.max_joined_circuits = 100

        # Maximum number of seconds that a circuit should exist
        self.max_time = 10 * 60
        # Maximum number of seconds that an introduction point should exist
        self.max_time_ip = 24 * 60 * 60
        # Maximum number of seconds before a circuit is considered inactive (and is removed)
        self.max_time_inactive = 20
        self.max_traffic = 250 * 1024 * 1024

        # Maximum number of seconds circuit creation is allowed to take. Within this time period, the unverified hop
        # of the circuit can still change in case it is unresponsive.
        self.circuit_timeout = 60
        # Maximum number of seconds that a hop allows us to change the next hop
        self.unstable_timeout = 60
        # Maximum number of seconds adding a single hop to a circuit is allowed to take.
        self.next_hop_timeout = 10

        self.swarm_lookup_interval = 30
        self.swarm_connection_limit = 15

        # We have a small delay when removing circuits/relays/exit nodes. This is to allow some post-mortem data
        # to flow over the circuit (i.e. bandwidth payouts to intermediate nodes in a circuit).
        self.remove_tunnel_delay = 5

        self.peer_flags = {PEER_FLAG_RELAY, PEER_FLAG_SPEED_TEST}

        # Maximum number of relay_early cells that are allowed to pass a relay.
        self.max_relay_early = 8

    @classmethod
    def from_dict(cls, d):
        result = cls()
        for k, v in d.items():
            setattr(result, k, v)
        return result


class TunnelCommunity(Community):

    version = b'\x02'
    community_id = unhexlify('81ded07332bdc775aa5a46f96de9f8f390bbc9f3')

    def __init__(self, *args, **kwargs):
        settings = kwargs.pop('settings', TunnelSettings())
        if isinstance(settings, dict):
            settings = TunnelSettings.from_dict(settings)
        self.settings = settings
        self.dht_provider = kwargs.pop('dht_provider', None)

        super(TunnelCommunity, self).__init__(*args, **kwargs)

        self.request_cache = RequestCache()
        self.decode_map_private = {}

        # Messages that can arrive from the socket
        self.add_message_handler(CellPayload, self.on_cell)
        self.add_message_handler(DestroyPayload, self.on_destroy)

        # Messages that can arrive from a circuit (i.e., they are wrapped in a cell)
        self.add_cell_handler(DataPayload, self.on_data)
        self.add_cell_handler(CreatePayload, self.on_create)
        self.add_cell_handler(CreatedPayload, self.on_created)
        self.add_cell_handler(ExtendPayload, self.on_extend)
        self.add_cell_handler(ExtendedPayload, self.on_extended)
        self.add_cell_handler(PingPayload, self.on_ping)
        self.add_cell_handler(PongPayload, self.on_pong)
        self.add_cell_handler(TestRequestPayload, self.on_test_request)
        self.add_cell_handler(TestResponsePayload, self.on_test_response)

        self.circuits = {}
        self.directions = {}
        self.relay_from_to = {}
        self.relay_session_keys = {}
        self.exit_sockets = {}
        self.circuits_needed = defaultdict(int)
        self.num_hops_by_downloads = defaultdict(int)  # Keeps track of the number of hops required by downloads
        self.candidates = {}  # Keeps track of the candidates that want to be a relay/exit node

        self.crypto = self.settings.crypto

        self.logger.info("Exit settings: BT=%s, IPv8=%s",
                         PEER_FLAG_EXIT_BT in self.settings.peer_flags,
                         PEER_FLAG_EXIT_IPV8 in self.settings.peer_flags)

        self.crypto.initialize(self.my_peer.key)

        if isinstance(self.endpoint, TunnelEndpoint):
            self.endpoint.set_tunnel_community(self)
            self.endpoint.set_anonymity(self._prefix, False)

        self.register_task("do_circuits", self.do_circuits, interval=5, delay=0)
        self.register_task("do_ping", self.do_ping, interval=PING_INTERVAL)

    async def unload(self):
        # Remove all circuits/relays/exitsockets
        for circuit_id in list(self.circuits.keys()):
            self.remove_circuit(circuit_id, 'unload', remove_now=True, destroy=DESTROY_REASON_SHUTDOWN)
        for circuit_id in list(self.relay_from_to.keys()):
            self.remove_relay(circuit_id, 'unload', remove_now=True, destroy=DESTROY_REASON_SHUTDOWN, both_sides=False)
        for circuit_id in list(self.exit_sockets.keys()):
            self.remove_exit_socket(circuit_id, 'unload', remove_now=True, destroy=DESTROY_REASON_SHUTDOWN)

        await self.request_cache.shutdown()

        await super(TunnelCommunity, self).unload()

    def get_serializer(self):
        serializer = super().get_serializer()
        serializer.add_packer('flags', Flags())
        return serializer

    def add_cell_handler(self, payload_cls, handler):
        self.decode_map_private[payload_cls.msg_id] = handler

    def _generate_circuit_id(self):
        circuit_id = random.getrandbits(32)

        # Prevent collisions.
        while circuit_id in self.circuits:
            circuit_id = random.getrandbits(32)

        return circuit_id

    def do_circuits(self):
        for circuit_length, num_circuits in self.circuits_needed.items():
            num_to_build = max(0, num_circuits - len(self.find_circuits(state=None, hops=circuit_length)))
            self.logger.info("Want %d data circuits of length %d", num_to_build, circuit_length)
            for _ in range(num_to_build):
                if not self.create_circuit(circuit_length):
                    self.logger.info("circuit creation of %d circuits failed, no need to continue", num_to_build)
                    break
        self.do_remove()

    def build_tunnels(self, hops):
        if hops > 0:
            self.circuits_needed[hops] = max(self.settings.max_circuits,
                                             min(self.settings.max_circuits, self.circuits_needed.get(hops, 0) + 1))
            self.do_circuits()

    def tunnels_ready(self, hops):
        if hops > 0 and self.circuits_needed.get(hops, 0):
            return len(self.find_circuits(hops=hops)) / float(self.circuits_needed[hops])
        return 1.0

    def do_remove(self):
        # Remove circuits that are inactive / are too old / have transferred too many bytes.
        for circuit_id, circuit in list(self.circuits.items()):
            if circuit.state == CIRCUIT_STATE_READY and \
               circuit.last_activity < time.time() - self.settings.max_time_inactive:
                self.remove_circuit(circuit_id, 'no activity')
            elif circuit.creation_time < time.time() - self.get_max_time(circuit_id):
                self.remove_circuit(circuit_id, 'too old')
            elif circuit.bytes_up + circuit.bytes_down > self.settings.max_traffic:
                self.remove_circuit(circuit_id, 'traffic limit exceeded')

        # Remove relays that are inactive / have transferred too many bytes.
        for circuit_id, relay in list(self.relay_from_to.items()):
            if relay.last_activity < time.time() - self.settings.max_time_inactive:
                self.remove_relay(circuit_id, 'no activity', both_sides=False)
            elif relay.bytes_up + relay.bytes_down > self.settings.max_traffic:
                self.remove_relay(circuit_id, 'traffic limit exceeded', both_sides=False)

        # Remove exit sockets that are too old / have transferred too many bytes.
        for circuit_id, exit_socket in list(self.exit_sockets.items()):
            if exit_socket.last_activity < time.time() - self.settings.max_time_inactive:
                self.remove_exit_socket(circuit_id, 'no activity')
            elif exit_socket.creation_time < time.time() - self.get_max_time(circuit_id):
                self.remove_exit_socket(circuit_id, 'too old')
            elif exit_socket.bytes_up + exit_socket.bytes_down > self.settings.max_traffic:
                self.remove_exit_socket(circuit_id, 'traffic limit exceeded')

        # Remove candidates that are not returned as verified peers
        current_peers = self.get_peers()
        for peer in list(self.candidates):
            if peer not in current_peers:
                self.candidates.pop(peer)

    def get_candidates(self, *requested_flags):
        return [peer for peer, flags in self.candidates.items()
                if set(requested_flags) <= set(flags) and self.crypto.is_key_compatible(peer.public_key)]

    def get_max_time(self, circuit_id):
        return self.settings.max_time

    def find_circuits(self, ctype=CIRCUIT_TYPE_DATA, state=CIRCUIT_STATE_READY, exit_flags=None, hops=None):
        return [c for c in self.circuits.values()
                if (state is None or c.state == state)
                and (ctype is None or c.ctype == ctype)
                and (exit_flags is None or set(exit_flags) <= set(c.exit_flags))
                and (hops is None or hops == c.goal_hops)]

    def create_circuit(self, goal_hops, ctype=CIRCUIT_TYPE_DATA, exit_flags=None, required_exit=None, info_hash=None):
        self.logger.info("Creating a new circuit of length %d (type: %s)", goal_hops, ctype)

        # Determine the last hop
        if not required_exit:
            if exit_flags is not None:
                exit_candidates = self.get_candidates(*exit_flags)
            elif ctype == CIRCUIT_TYPE_DATA:
                exit_candidates = self.get_candidates(PEER_FLAG_EXIT_BT)
            elif ctype == CIRCUIT_TYPE_IP_SEEDER:
                # For introduction points we prefer exit nodes, but perhaps a relay peer would also suffice..
                exit_candidates = self.get_candidates(PEER_FLAG_EXIT_BT) \
                    or self.get_candidates(PEER_FLAG_EXIT_IPV8) \
                    or self.get_candidates(PEER_FLAG_RELAY)
            else:
                # For exit nodes that don't exit actual data, we prefer relay candidates,
                # but we also consider exit candidates.
                exit_candidates = self.get_candidates(PEER_FLAG_RELAY) or self.get_candidates(PEER_FLAG_EXIT_BT)

            if not exit_candidates:
                self.logger.info("Could not create circuit, no available exit-nodes")
                return

            required_exit = random.choice(exit_candidates)

        # Determine the first hop
        if goal_hops == 1 and required_exit:
            # If the number of hops is 1, it should immediately be the required_exit hop.
            self.logger.info("First hop is required exit")
            possible_first_hops = [required_exit]
        else:
            self.logger.info("Look for a first hop that is not an exit node and is not used before")
            # First build a list of hops, then filter the list. Avoids issues when create_circuit is called
            # from a different thread (caused by circuit.peer being reset to None).
            first_hops = [c.peer for c in self.circuits.values()]
            first_hops = {h for h in first_hops if h}
            relay_candidates = self.get_candidates(PEER_FLAG_RELAY)
            possible_first_hops = [c for c in relay_candidates if c not in first_hops and c != required_exit]

        if not possible_first_hops:
            self.logger.info("Could not create circuit, no first hop available")
            return

        # Finally, construct the Circuit object and send the CREATE message
        circuit_id = self._generate_circuit_id()
        self.circuits[circuit_id] = circuit = Circuit(circuit_id, goal_hops, ctype, required_exit, info_hash)
        self.send_initial_create(circuit, possible_first_hops,
                                 self.settings.circuit_timeout // self.settings.next_hop_timeout)

        return circuit

    def send_initial_create(self, circuit, candidate_list, max_tries):
        first_hop = random.choice(candidate_list)
        alt_first_hops = [c for c in candidate_list if c != first_hop]

        circuit.unverified_hop = Hop(first_hop, flags=self.candidates.get(first_hop))
        circuit.unverified_hop.dh_secret, circuit.unverified_hop.dh_first_part = self.crypto.generate_diffie_secret()

        self.logger.info("Adding first hop %s:%d to circuit %d", *(first_hop.address + (circuit.circuit_id,)))

        cache = RetryRequestCache(self, circuit, alt_first_hops, max_tries - 1,
                                  self.send_initial_create, self.settings.next_hop_timeout)
        self.request_cache.add(cache)

        self.send_cell(first_hop, CreatePayload(circuit.circuit_id,
                                                cache.number,
                                                self.my_peer.public_key.key_to_bin(),
                                                circuit.unverified_hop.dh_first_part))

    @task
    async def remove_circuit(self, circuit_id, additional_info='', remove_now=False, destroy=False):
        """
        Remove a circuit and return a deferred that fires when all data associated with the circuit is destroyed.
        Optionally send a destroy message.
        """
        circuit_to_remove = self.circuits.get(circuit_id, None)
        if circuit_to_remove is None:
            self.logger.warning('Cannot remove unknown circuit %d', circuit_id)
            return

        self.logger.info("Removing %s circuit %d %s", circuit_to_remove.ctype, circuit_id, additional_info)

        if destroy:
            self.destroy_circuit(circuit_to_remove, reason=destroy)

        circuit_to_remove.close(additional_info)

        if not remove_now or self.settings.remove_tunnel_delay > 0:
            await sleep(self.settings.remove_tunnel_delay)

        circuit = self.circuits.pop(circuit_id, None)
        if circuit:
            self.logger.info("Removed circuit %d " + additional_info, circuit_id)

        # Clean up the directions dictionary
        self.directions.pop(circuit_id, None)

    @task
    async def remove_relay(self, circuit_id, additional_info='', remove_now=False, destroy=False,
                           got_destroy_from=None, both_sides=True):
        """
        Remove a relay and all information associated with the relay. Return the relays that have been removed.
        """
        to_remove = [circuit_id]
        if both_sides:
            # Find other side of relay
            for k, v in self.relay_from_to.items():
                if circuit_id == v.circuit_id:
                    to_remove.append(k)

        # Send destroy
        if destroy:
            self.destroy_relay(to_remove, got_destroy_from=got_destroy_from, reason=destroy)

        if not remove_now or self.settings.remove_tunnel_delay > 0:
            await sleep(self.settings.remove_tunnel_delay)

        removed_relays = []
        for cid in to_remove:
            # Remove the relay
            self.logger.info("Removing relay %d %s", cid, additional_info)

            relay = self.relay_from_to.pop(cid, None)
            if relay:
                removed_relays.append(relay)

            # Remove old session key
            self.relay_session_keys.pop(cid, None)

            # Clean directions dictionary
            self.directions.pop(cid, None)

        return removed_relays

    @task
    async def remove_exit_socket(self, circuit_id, additional_info='', remove_now=False, destroy=False):
        """
        Remove an exit socket. Send a destroy message if necessary.
        """
        exit_socket_to_destroy = self.exit_sockets.get(circuit_id, None)
        if exit_socket_to_destroy and destroy:
            self.destroy_exit_socket(exit_socket_to_destroy, reason=destroy)

        if not remove_now or self.settings.remove_tunnel_delay > 0:
            await sleep(self.settings.remove_tunnel_delay)

        exit_socket = self.exit_sockets.pop(circuit_id, None)
        if exit_socket:
            # Close socket
            if exit_socket.enabled:
                self.logger.info("Removing exit socket %d %s", circuit_id, additional_info)
                await exit_socket.close()
                # Remove old session key
                self.relay_session_keys.pop(circuit_id, None)
            await exit_socket.shutdown_task_manager()
        return exit_socket

    def destroy_circuit(self, circuit, reason=0):
        sock_addr = circuit.peer.address
        self.send_destroy(sock_addr, circuit.circuit_id, reason)
        self.logger.info("destroy_circuit %s %s", circuit.circuit_id, sock_addr)

    def destroy_relay(self, circuit_ids, reason=0, got_destroy_from=None):
        relays = {cid_from: (self.relay_from_to[cid_from].circuit_id,
                             self.relay_from_to[cid_from].peer.address) for cid_from in circuit_ids
                  if cid_from in self.relay_from_to}

        if got_destroy_from and got_destroy_from not in relays.values():
            self.logger.error("%s not allowed send destroy for circuit %s", *reversed(got_destroy_from))
            return

        for cid_from, (cid_to, sock_addr) in relays.items():
            self.logger.info("Found relay %s -> %s (%s)", cid_from, cid_to, sock_addr)
            if (cid_to, sock_addr) != got_destroy_from:
                self.send_destroy(sock_addr, cid_to, reason)
                self.logger.info("Fw destroy to %s %s", cid_to, sock_addr)

    def destroy_exit_socket(self, exit_socket, reason=0):
        sock_addr = exit_socket.peer.address
        self.send_destroy(sock_addr, exit_socket.circuit_id, reason)
        self.logger.info("Destroy_exit_socket %s %s", exit_socket.circuit_id, sock_addr)

    def send_cell(self, peer, payload):
        circuit_id = payload.circuit_id
        message = self.serializer.pack_serializable(payload)[4:]
        cell = CellPayload(circuit_id, pack('!B', payload.msg_id) + message)

        cell.plaintext = payload.msg_id in NO_CRYPTO_PACKETS
        if circuit_id in self.circuits:
            circuit = self.circuits[circuit_id]
            cell.relay_early = payload.msg_id == 4 or circuit.relay_early_count < self.settings.max_relay_early
            if cell.relay_early:
                circuit.relay_early_count += 1

        try:
            cell.encrypt(self.crypto, self.circuits.get(circuit_id), self.relay_session_keys.get(circuit_id))
        except CryptoException as e:
            self.logger.warning(str(e))
            return
        packet = cell.to_bin(self._prefix)
        packet_len = self.send_packet(peer, packet)

        tunnel_obj = self.circuits.get(circuit_id) or self.relay_from_to.get(circuit_id)
        if tunnel_obj:
            tunnel_obj.bytes_up += packet_len

    def send_data(self, peer, circuit_id, dest_address, source_address, data):
        payload = DataPayload(circuit_id, dest_address, source_address, data)
        return self.send_cell(peer, payload)

    def send_packet(self, peer, packet):
        self.endpoint.send(peer if isinstance(peer, tuple) else peer.address, packet)
        return len(packet)

    def send_destroy(self, peer, circuit_id, reason):
        auth = BinMemberAuthenticationPayload(self.my_peer.public_key.key_to_bin())
        payload = DestroyPayload(circuit_id, reason)
        packet = self._ez_pack(self._prefix, DestroyPayload.msg_id, [auth, payload])
        self.send_packet(peer, packet)

    def relay_cell(self, cell):
        next_relay = self.relay_from_to[cell.circuit_id]
        if cell.plaintext:
            self.logger.warning('Dropping cell (cell not encrypted)')
            return
        if cell.relay_early and next_relay.relay_early_count >= self.settings.max_relay_early:
            self.logger.warning('Dropping cell (too many relay_early cells)')
            return

        try:
            if next_relay.rendezvous_relay:
                cell.decrypt(self.crypto, relay_session_keys=self.relay_session_keys[cell.circuit_id])
                cell.encrypt(self.crypto, relay_session_keys=self.relay_session_keys[next_relay.circuit_id])
                cell.relay_early = False
            else:
                direction = self.directions[cell.circuit_id]
                if direction == ORIGINATOR:
                    cell.encrypt(self.crypto, relay_session_keys=self.relay_session_keys[cell.circuit_id])
                elif direction == EXIT_NODE:
                    cell.decrypt(self.crypto, relay_session_keys=self.relay_session_keys[cell.circuit_id])
        except CryptoException as e:
            self.logger.warning(str(e))
            return

        cell.circuit_id = next_relay.circuit_id
        packet = cell.to_bin(self._prefix)
        next_relay.bytes_up += self.send_packet(next_relay.peer, packet)
        next_relay.relay_early_count += 1

    def _ours_on_created_extended(self, circuit, payload):
        hop = circuit.unverified_hop

        try:
            shared_secret = self.crypto.verify_and_generate_shared_secret(hop.dh_secret, payload.key,
                                                                          payload.auth, hop.public_key.key.pk)
            hop.session_keys = self.crypto.generate_session_keys(shared_secret)

        except CryptoException:
            self.remove_circuit(circuit.circuit_id, "error while verifying shared secret")
            return

        circuit.unverified_hop = None
        circuit.add_hop(hop)

        if circuit.state == CIRCUIT_STATE_EXTENDING:
            candidate_list_enc = payload.candidate_list_enc
            candidate_list_bin = self.crypto.decrypt_str(candidate_list_enc,
                                                         hop.session_keys[EXIT_NODE],
                                                         hop.session_keys[EXIT_NODE_SALT])
            candidate_list, _ = self.serializer.unpack('varlenH-list', candidate_list_bin)

            cache = self.request_cache.pop("retry", payload.identifier)
            self.send_extend(circuit, candidate_list, cache.max_tries if cache else 1)

        elif circuit.state == CIRCUIT_STATE_READY:
            self.request_cache.pop("retry", payload.identifier)

    def send_extend(self, circuit, candidate_list, max_tries):
        ignore_candidates = [hop.node_public_key for hop in circuit.hops] + [self.my_peer.public_key.key_to_bin()]
        if circuit.required_exit:
            ignore_candidates.append(circuit.required_exit.public_key.key_to_bin())

        become_exit = circuit.goal_hops - 1 == len(circuit.hops)
        if become_exit and circuit.required_exit:
            # Set the required exit according to the circuit setting (e.g. for linking e2e circuits)
            extend_hop_public_bin = circuit.required_exit.public_key.key_to_bin()
            extend_hop_addr = circuit.required_exit.address

        else:
            # The next candidate is chosen from the returned list of possible candidates
            for ignore_candidate in ignore_candidates:
                if ignore_candidate in candidate_list:
                    candidate_list.remove(ignore_candidate)

            for i in range(len(candidate_list) - 1, -1, -1):
                public_key = self.crypto.key_from_public_bin(candidate_list[i])
                if not self.crypto.is_key_compatible(public_key):
                    candidate_list.pop(i)

            extend_hop_public_bin = next(iter(candidate_list), None)
            extend_hop_addr = ('0.0.0.0', 0)

        if extend_hop_public_bin:
            extend_hop_public_key = self.crypto.key_from_public_bin(extend_hop_public_bin)
            circuit.unverified_hop = Hop(Peer(extend_hop_public_key),
                                         flags=self.candidates.get(Peer(extend_hop_public_bin)))
            circuit.unverified_hop.dh_secret, circuit.unverified_hop.dh_first_part = \
                self.crypto.generate_diffie_secret()

            self.logger.info("Extending circuit %d with %s", circuit.circuit_id, hexlify(extend_hop_public_bin))

            # Only retry if we are allowed to use another node
            if not become_exit or not circuit.required_exit:
                alt_candidates = [c for c in candidate_list if c != extend_hop_public_bin]
            else:
                alt_candidates = []

            cache = RetryRequestCache(self, circuit, alt_candidates, max_tries - 1,
                                      self.send_extend, self.settings.next_hop_timeout)
            self.request_cache.add(cache)

            self.send_cell(circuit.peer, ExtendPayload(circuit.circuit_id,
                                                       cache.number,
                                                       circuit.unverified_hop.node_public_key,
                                                       circuit.unverified_hop.dh_first_part,
                                                       extend_hop_addr))

        else:
            self.remove_circuit(circuit.circuit_id, "no candidates to extend")

    def extract_peer_flags(self, extra_bytes):
        if not extra_bytes:
            return []

        payload, _ = self.serializer.unpack_serializable(ExtraIntroductionPayload, extra_bytes)
        return payload.flags

    def introduction_request_callback(self, peer, dist, payload):
        self.candidates[peer] = self.extract_peer_flags(payload.extra_bytes)

    def introduction_response_callback(self, peer, dist, payload):
        self.candidates[peer] = self.extract_peer_flags(payload.extra_bytes)

    def create_introduction_request(self, socket_address, extra_bytes=b'', new_style=False):
        extra_payload = ExtraIntroductionPayload(self.settings.peer_flags)
        extra_bytes = self.serializer.pack_serializable(extra_payload)
        return super().create_introduction_request(socket_address, extra_bytes, new_style)

    def create_introduction_response(self, lan_socket_address, socket_address, identifier,
                                     introduction=None, extra_bytes=b'', prefix=None, new_style=False):
        extra_payload = ExtraIntroductionPayload(self.settings.peer_flags)
        extra_bytes = self.serializer.pack_serializable(extra_payload)
        return super(TunnelCommunity, self).create_introduction_response(lan_socket_address, socket_address,
                                                                         identifier, introduction, extra_bytes,
                                                                         prefix, new_style)

    def on_cell(self, source_address, data):
        cell = CellPayload.from_bin(data)
        circuit_id = cell.circuit_id

        if circuit_id in self.relay_from_to:
            next_relay = self.relay_from_to[circuit_id]
            this_relay = self.relay_from_to.get(next_relay.circuit_id, None)
            if this_relay:
                this_relay.beat_heart()
                this_relay.bytes_down += len(data)
            self.logger.debug("Relaying cell from circuit %d to %d", circuit_id, next_relay.circuit_id)
            self.relay_cell(cell)
            return

        circuit = self.circuits.get(circuit_id, None)
        try:
            cell.decrypt(self.crypto, circuit=circuit, relay_session_keys=self.relay_session_keys.get(circuit_id))
        except CryptoException as e:
            self.logger.debug(str(e))
            if circuit:
                self.send_destroy(circuit.peer, circuit_id, 0)
            return
        self.logger.debug("Got cell(%s) from circuit %d (sender %s, receiver %s)",
                          cell.message[0], circuit_id, source_address, self.my_peer)

        if (not cell.relay_early and cell.message[0] == 4) or self.settings.max_relay_early <= 0:
            self.logger.info('Dropping cell (missing or unexpected relay_early flag)')
            return
        if cell.plaintext and cell.message[0] not in NO_CRYPTO_PACKETS:
            self.logger.warning('Dropping cell (only create/created can have plaintext flag set)')
            return

        self.on_packet_from_circuit(source_address, cell.unwrap(self._prefix), circuit_id)

        if circuit:
            circuit.beat_heart()
            circuit.bytes_down += len(data)

    def on_packet_from_circuit(self, source_address, data, circuit_id):
        if self._prefix != data[:22]:
            return
        msg_id = data[22]
        if msg_id in self.decode_map_private:
            try:
                handler = self.decode_map_private[msg_id]
                result = handler(source_address, data, circuit_id)
                if iscoroutine(result):
                    self.register_anonymous_task('on_packet_from_circuit', ensure_future(result), ignore=(Exception,))
            except Exception:
                self.logger.error("Exception occurred while handling packet!\n"
                                  + ''.join(format_exception(*sys.exc_info())))

    async def should_join_circuit(self, create_payload, previous_node_address):
        """
        Check whether we should join a circuit.
        Returns a deferred that fires with a boolean.
        """
        if self.settings.max_joined_circuits <= len(self.relay_from_to) + len(self.exit_sockets):
            self.logger.warning("Too many relays (%d)", (len(self.relay_from_to) + len(self.exit_sockets)))
            return False
        return True

    def join_circuit(self, create_payload, previous_node_address):
        """
        Actively join a circuit and send a created message back
        """
        circuit_id = create_payload.circuit_id

        self.directions[circuit_id] = EXIT_NODE
        self.logger.info('We joined circuit %d with neighbour %s', circuit_id, previous_node_address)

        shared_secret, key, auth = self.crypto.generate_diffie_shared_secret(create_payload.key)
        self.relay_session_keys[circuit_id] = self.crypto.generate_session_keys(shared_secret)

        peers_list = [peer for peer in self.get_candidates(PEER_FLAG_RELAY)
                      if peer not in self.get_candidates(PEER_FLAG_EXIT_BT)][:4]
        peers_keys = {c.public_key.key_to_bin(): c for c in peers_list}

        peer = Peer(create_payload.node_public_key, previous_node_address)
        self.request_cache.add(CreatedRequestCache(self, circuit_id, peer, peers_keys, self.settings.unstable_timeout))
        self.exit_sockets[circuit_id] = TunnelExitSocket(circuit_id, peer, self)

        candidate_list_bin = self.serializer.pack('varlenH-list', list(peers_keys.keys()))
        candidate_list_enc = self.crypto.encrypt_str(candidate_list_bin,
                                                     *self.crypto.get_session_keys(self.relay_session_keys[circuit_id],
                                                                                   EXIT_NODE))
        self.send_cell(Peer(create_payload.node_public_key, previous_node_address),
                       CreatedPayload(circuit_id, create_payload.identifier, key, auth, candidate_list_enc))

    @unpack_cell(CreatePayload)
    async def on_create(self, source_address, payload, _):
        if not self.settings.peer_flags:
            self.logger.warning("Ignoring create for circuit %d", payload.circuit_id)
            return
        if self.request_cache.has("created", payload.circuit_id):
            self.logger.warning("Already have a request for circuit %d", payload.circuit_id)
            return

        result = await self.should_join_circuit(payload, source_address)
        if result:
            self.join_circuit(payload, source_address)
        else:
            self.logger.warning("We're not joining circuit with ID %s", payload.circuit_id)

    @unpack_cell(CreatedPayload)
    def on_created(self, source_address, payload, _):
        circuit_id = payload.circuit_id
        self.directions[circuit_id] = ORIGINATOR

        if self.request_cache.has("create", payload.identifier):
            request = self.request_cache.pop("create", payload.identifier)

            self.logger.info("Got CREATED message forward as EXTENDED to origin.")

            self.relay_from_to[request.to_circuit_id] = relay = RelayRoute(request.from_circuit_id, request.peer)
            self.relay_from_to[request.from_circuit_id] = RelayRoute(request.to_circuit_id, request.to_peer)
            self.relay_session_keys[request.to_circuit_id] = self.relay_session_keys[request.from_circuit_id]

            self.directions[request.from_circuit_id] = EXIT_NODE
            self.remove_exit_socket(request.from_circuit_id)

            self.send_cell(relay.peer,
                           ExtendedPayload(relay.circuit_id, request.extend_identifier,
                                           payload.key, payload.auth, payload.candidate_list_enc))
        elif self.request_cache.has("retry", payload.identifier):
            circuit = self.circuits[circuit_id]
            self._ours_on_created_extended(circuit, payload)
        else:
            self.logger.warning("Received unexpected created for circuit %d", payload.circuit_id)

    @unpack_cell(ExtendPayload)
    async def on_extend(self, source_address, payload, _):
        if PEER_FLAG_RELAY not in self.settings.peer_flags:
            self.logger.warning("Ignoring create for circuit %d", payload.circuit_id)
            return

        if not self.request_cache.has("created", payload.circuit_id):
            self.logger.warning("Received unexpected extend for circuit %d", payload.circuit_id)
            return

        circuit_id = payload.circuit_id
        # Leave the RequestCache in case the circuit owner wants to reuse the tunnel for a different next-hop
        request = self.request_cache.get("created", circuit_id)
        if payload.node_addr == ('0.0.0.0', 0) and payload.node_public_key not in request.candidates:
            self.logger.warning("Node public key not in request candidates and no ip specified")
            return

        if payload.node_public_key in request.candidates:
            extend_candidate = request.candidates[payload.node_public_key]
        else:
            extend_candidate = self.network.get_verified_by_public_key_bin(payload.node_public_key)
            if not extend_candidate:
                extend_candidate = Peer(payload.node_public_key, payload.node_addr)

            # Ensure that we are able to contact this peer
            if extend_candidate.last_response + 57.5 < time.time():
                await self.dht_peer_lookup(extend_candidate.mid, peer=extend_candidate)

        self.logger.info("On_extend send CREATE for circuit (%s, %d) to %s:%d", source_address,
                         circuit_id, *extend_candidate.address)

        to_circuit_id = self._generate_circuit_id()

        if circuit_id in self.circuits:
            candidate = self.circuits[circuit_id].peer
        elif circuit_id in self.exit_sockets:
            candidate = self.exit_sockets[circuit_id].peer
        elif circuit_id in self.relay_from_to:
            candidate = self.relay_from_to[circuit_id].peer
        else:
            self.logger.error("Got extend for unknown source circuit_id")
            return

        self.logger.info("Extending circuit, got candidate with IP %s:%d from cache", *extend_candidate.address)

        cache = CreateRequestCache(self, payload.identifier, to_circuit_id, circuit_id, candidate, extend_candidate)
        self.request_cache.add(cache)

        self.send_cell(extend_candidate,
                       CreatePayload(to_circuit_id, cache.number, self.my_peer.public_key.key_to_bin(), payload.key))

    @unpack_cell(ExtendedPayload)
    def on_extended(self, source_address, payload, _):
        if not self.request_cache.has("retry", payload.identifier):
            self.logger.warning("Received unexpected extended for circuit %s", payload.circuit_id)
            return

        circuit_id = payload.circuit_id
        circuit = self.circuits[circuit_id]
        self._ours_on_created_extended(circuit, payload)

    def on_raw_data(self, circuit, origin, data):
        """
        Handle data, coming from a specific circuit and origin.
        This method is usually implemented in subclasses of this community.
        """
        pass

    def on_data(self, sock_addr, data, _):
        payload, _ = self.serializer.unpack_serializable(DataPayload, data, offset=23)

        # If its our circuit, the messenger is the candidate assigned to that circuit and the DATA's destination
        # is set to the zero-address then the packet is from the outside world and addressed to us from.
        circuit_id = payload.circuit_id
        destination = payload.dest_address
        origin = payload.org_address
        data = payload.data

        self.logger.debug("Got data (%d) from %s", circuit_id, sock_addr)

        circuit = self.circuits.get(circuit_id, None)
        if circuit and origin and sock_addr == circuit.peer.address:
            circuit.beat_heart()

            e2e_data = circuit.ctype in [CIRCUIT_TYPE_RP_DOWNLOADER, CIRCUIT_TYPE_RP_SEEDER]
            if DataChecker.could_be_ipv8(data) and not e2e_data:
                if self._prefix == data[:22]:
                    self.logger.debug("Incoming packet meant for us")
                    self.on_packet_from_circuit(origin, data, circuit_id)
                    return

                if isinstance(self.endpoint, TunnelEndpoint):
                    self.logger.debug("Incoming packet meant for other community")
                    self.endpoint.notify_listeners((origin, data), from_tunnel=True)
                else:
                    self.logger.debug("Incoming packet meant for other community, dropping")
            else:
                # We probably received raw data, handle it
                self.on_raw_data(circuit, origin, data)

        # It is not our circuit so we got it from a relay, we need to EXIT it!
        else:
            self.logger.debug("Data for circuit %d exiting tunnel (%s)", circuit_id, destination)
            if destination != ('0.0.0.0', 0):
                self.exit_data(circuit_id, sock_addr, destination, data)
            else:
                self.logger.warning("Cannot exit data, destination is 0.0.0.0:0")

    @unpack_cell(PingPayload)
    def on_ping(self, source_address, payload, _):
        if not (payload.circuit_id in self.circuits
                or payload.circuit_id in self.exit_sockets
                or payload.circuit_id in self.relay_from_to):
            return

        exit_socket = self.exit_sockets.get(payload.circuit_id)
        if exit_socket:
            exit_socket.beat_heart()

        self.send_cell(source_address, PongPayload(payload.circuit_id, payload.identifier))
        self.logger.debug("Got ping from %s", source_address)

    @unpack_cell(PongPayload)
    def on_pong(self, source_address, payload, _):
        if not self.request_cache.has("ping", payload.identifier):
            self.logger.warning("Invalid ping circuit_id")
            return

        self.request_cache.pop("ping", payload.identifier)
        self.logger.debug("Got pong from %s", source_address)

    def do_ping(self, exclude=None):
        # Ping circuits. Pings are only sent to the first hop, subsequent hops will relay the ping.
        exclude = [] if exclude is None else exclude
        for circuit in list(self.circuits.values()):
            if circuit.state in [CIRCUIT_STATE_READY, CIRCUIT_STATE_EXTENDING] \
                    and circuit.circuit_id not in exclude \
                    and circuit.hops:
                cache = PingRequestCache(self)
                self.request_cache.add(cache)
                self.send_cell(circuit.peer, PingPayload(circuit.circuit_id, cache.number))

    @lazy_wrapper(DestroyPayload)
    def on_destroy(self, peer, payload):
        source_address = peer.address
        circuit_id = payload.circuit_id
        self.logger.info("Got destroy from %s for circuit %s", source_address, circuit_id)

        if circuit_id in self.relay_from_to:
            self.remove_relay(circuit_id, "got destroy", destroy=DESTROY_REASON_FORWARD,
                              got_destroy_from=(circuit_id, source_address))

        elif circuit_id in self.exit_sockets and source_address == self.exit_sockets[circuit_id].peer.address:
            self.logger.info("Got an exit socket %s %s", circuit_id, source_address)
            self.remove_exit_socket(circuit_id, f"got destroy with reason {payload.reason}")

        elif circuit_id in self.circuits and source_address == self.circuits[circuit_id].peer.address:
            self.logger.info("Got a circuit %s %s", circuit_id, source_address)
            self.remove_circuit(circuit_id, f"got destroy with reason {payload.reason}")

        else:
            self.logger.warning("Invalid or unauthorized destroy")

    def exit_data(self, circuit_id, sock_addr, destination, data):
        if circuit_id not in self.exit_sockets:
            self.logger.error("Dropping data packets with unknown circuit_id")
            return

        if not self.exit_sockets[circuit_id].enabled:
            # Check that we got the data from the correct IP.
            if sock_addr[0] == self.exit_sockets[circuit_id].peer.address[0]:
                self.exit_sockets[circuit_id].enable()
            else:
                self.logger.error("Dropping outbound relayed packet: IP's are %s != %s",
                                  str(sock_addr), str(self.exit_sockets[circuit_id].peer.address))
                return
        try:
            self.exit_sockets[circuit_id].sendto(data, destination)
        except Exception:
            self.logger.warning("Dropping data packets while exiting")

    async def dht_peer_lookup(self, mid, peer=None):
        if self.dht_provider:
            await self.dht_provider.peer_lookup(mid, peer)
        else:
            self.logger.error("Need a DHT provider to connect to a peer using the DHT")

    def send_test_request(self, circuit, request_size=0, response_size=0):
        cache = TestRequestCache(self, circuit)
        self.request_cache.add(cache)
        self.send_cell(circuit.peer,
                       TestRequestPayload(circuit.circuit_id, cache.number, response_size, os.urandom(request_size)))
        return cache.future

    def on_test_request(self, source_address, data, circuit_id):
        if PEER_FLAG_SPEED_TEST not in self.settings.peer_flags:
            self.logger.warning("Ignoring test-request from circuit %d", circuit_id)
            return

        payload, _ = self.serializer.unpack_serializable(TestRequestPayload, data, offset=23)
        exit_socket = self.exit_sockets.get(circuit_id)
        circuit = self.circuits.get(circuit_id)
        if not exit_socket and not (circuit and circuit.ctype in [CIRCUIT_TYPE_RP_SEEDER, CIRCUIT_TYPE_RP_DOWNLOADER]):
            self.logger.error("Dropping test-request with unknown circuit_id")
            return

        self.logger.debug("Got test-request (%d) from %s, replying with response", circuit_id, source_address)
        if exit_socket:
            exit_socket.beat_heart()
        self.send_cell(source_address,
                       TestResponsePayload(circuit_id, payload.identifier, os.urandom(payload.response_size)))

    def on_test_response(self, source_address, data, circuit_id):
        payload, _ = self.serializer.unpack_serializable(TestResponsePayload, data, offset=23)
        circuit = self.circuits.get(circuit_id)
        if not circuit:
            self.logger.error("Dropping test-response with unknown circuit_id")
            return
        if not self.request_cache.has("test-request", payload.identifier):
            self.logger.error("Dropping unexpected test-response")
            return

        self.logger.debug("Got test-response (%d) from %s", circuit_id, source_address)
        cache = self.request_cache.pop("test-request", payload.identifier)
        cache.future.set_result((payload.data, time.time() - cache.ts))



import base64
import json
import os
import urllib.parse

from ..REST.rest_base import RESTTestBase, partial_cls
from ...attestation.communication_manager import CommunicationChannel
from ...attestation.default_identity_formats import FORMATS
from ...attestation.identity.community import IdentityCommunity
from ...attestation.identity.manager import IdentityManager
from ...attestation.wallet.community import AttestationCommunity
from ...keyvault.crypto import ECCrypto


class TestIdentityEndpoint(RESTTestBase):
    """
    Class for testing the REST API of the IdentityEndpoint
    """

    async def setUp(self):
        super(TestIdentityEndpoint, self).setUp()

        self.pseudonym_directories = []
        identity_manager = IdentityManager(u":memory:")

        await self.initialize([partial_cls(IdentityCommunity, identity_manager=identity_manager,
                                           working_directory=':memory:')], 2)

    async def create_node(self, *args, **kwargs):
        """
        We load each node i with a pseudonym `my_peer{i}`, which is the default IPv8 `my_peer` key.
        """
        ipv8 = await super(TestIdentityEndpoint, self).create_node(*args, **kwargs)
        temp_dir = self.temporary_directory()
        key_file_name = 'my_peer' + str(len(self.pseudonym_directories))
        with open(os.path.join(temp_dir, key_file_name), 'wb') as f:
            f.write(ipv8.my_peer.key.key_to_bin())
        communication_manager = ipv8.rest_manager.root_endpoint.endpoints['/identity'].communication_manager
        communication_manager.working_directory = ':memory:'
        communication_manager.pseudonym_folder = temp_dir

        identity_overlay = ipv8.get_overlay(IdentityCommunity)
        attestation_overlay = AttestationCommunity(identity_overlay.my_peer, identity_overlay.endpoint,
                                                   identity_overlay.network, working_directory=':memory:')
        channel = CommunicationChannel(attestation_overlay, identity_overlay)

        communication_manager.channels[ipv8.my_peer.public_key.key_to_bin()] = channel
        communication_manager.name_to_channel[key_file_name] = channel
        self.pseudonym_directories.append(temp_dir)
        return ipv8

    async def introduce_pseudonyms(self):
        all_interfaces = []
        for node in self.nodes:
            all_interfaces.append([overlay.endpoint.wan_address for overlay in node.overlays
                                   if isinstance(overlay, IdentityCommunity)])
        for i in range(len(self.nodes)):
            other_addresses = list(range(len(self.nodes)))
            other_addresses.remove(i)
            for j in other_addresses:
                for overlay in self.node(i).overlays:
                    if isinstance(overlay, IdentityCommunity):
                        for address in all_interfaces[j]:
                            overlay.walk_to(address)
            await self.deliver_messages()

    async def wait_for(self, *args, **kwargs):
        output = []
        while not output:
            output = await self.make_request(*args)
            await self.deliver_messages()
        return output

    async def wait_for_requests(self, *args, **kwargs):
        requests = []
        while not requests:
            output = await self.wait_for(*args)
            if 'requests' in output:
                requests = output['requests']
            else:
                await self.deliver_messages()
        return requests

    async def test_list_pseudonyms_empty(self):
        """
        Check that we do not start with any pseudonyms.
        """
        await self.make_request(self.node(0), 'identity/my_peer0/remove', 'get')
        result = await self.make_request(self.node(0), 'identity', 'get')

        self.assertDictEqual({'names': []}, result)

    async def test_list_schemas(self):
        """
        Check that the endpoint reports the available schemas correctly.
        """
        schemas = await self.make_request(self.node(0), 'identity/test_pseudonym/schemas', 'get')

        self.assertSetEqual(set(FORMATS.keys()), set(schemas['schemas']))

    async def test_list_pseudonyms_one(self):
        """
        Check that a loaded pseudonym is reported as such.
        """
        result = await self.make_request(self.node(0), 'identity', 'get')

        self.assertDictEqual({'names': ['my_peer0']}, result)

    async def test_list_pseudonyms_many(self):
        """
        Check that all loaded pseudonyms are reported as such.
        """
        pseudonyms = ['test_pseudonym1', 'test_pseudonym2', 'test_pseudonym3', 'test_pseudonym4']
        for pseudonym in pseudonyms:
            await self.make_request(self.node(0), f'identity/{pseudonym}/schemas', 'get')

        result = await self.make_request(self.node(0), 'identity', 'get')

        self.assertSetEqual(set(pseudonyms) | {'my_peer0'}, set(result['names']))

    async def test_list_public_key_one(self):
        """
        Check that we retrieve the pseudonym public key correctly.
        """
        result = await self.make_request(self.node(0), 'identity/test_pseudonym/public_key', 'get')
        decoded_public_key = base64.b64decode(result['public_key'])

        # This should have made the `test_pseudonym` private key file (corresponding to the reported public key).
        key_file = os.path.join(self.pseudonym_directories[0], 'test_pseudonym')
        with open(key_file, 'rb') as key_file_handle:
            private_key = ECCrypto().key_from_private_bin(key_file_handle.read())

        self.assertEqual(private_key.pub().key_to_bin(), decoded_public_key)

    async def test_list_public_key_many(self):
        """
        Check that we retrieve the pseudonym public key correctly.
        """
        pseudonyms = ['test_pseudonym1', 'test_pseudonym2', 'test_pseudonym3', 'test_pseudonym4']
        # Make sure all pseudonyms exist before querying their keys.
        # This is not necessary, but has the highest chance of exposing failures.
        for pseudonym in pseudonyms:
            await self.make_request(self.node(0), f'identity/{pseudonym}/schemas', 'get')
        for pseudonym in pseudonyms:
            result = await self.make_request(self.node(0), f'identity/{pseudonym}/public_key', 'get')
            decoded_public_key = base64.b64decode(result['public_key'])

            key_file = os.path.join(self.pseudonym_directories[0], pseudonym)
            with open(key_file, 'rb') as key_file_handle:
                private_key = ECCrypto().key_from_private_bin(key_file_handle.read())

            self.assertEqual(private_key.pub().key_to_bin(), decoded_public_key)

    async def test_list_peers(self):
        """
        Check if peers are correctly listed.
        """
        result1 = await self.make_request(self.node(0), 'identity/my_peer0/peers', 'get')
        result2 = await self.make_request(self.node(1), 'identity/my_peer1/peers', 'get')

        self.assertListEqual([], result1['peers'])
        self.assertListEqual([], result2['peers'])

        await self.introduce_pseudonyms()

        result1 = await self.make_request(self.node(0), 'identity/my_peer0/peers', 'get')
        result2 = await self.make_request(self.node(1), 'identity/my_peer1/peers', 'get')

        self.assertEqual(1, len(result1['peers']))
        self.assertEqual(1, len(result2['peers']))

    async def test_list_unload(self):
        """
        Check if a pseudonym stops communicating on unload.
        """
        await self.make_request(self.node(0), 'identity/my_peer0/unload', 'get')
        await self.make_request(self.node(1), 'identity/my_peer1/unload', 'get')

        self.assertListEqual([], self.node(0).overlays)
        self.assertListEqual([], self.node(1).overlays)

    async def test_list_credentials_empty(self):
        """
        Check that we retrieve credentials correctly, if none exist.
        """
        result = await self.make_request(self.node(0), 'identity/my_peer0/credentials', 'get')

        self.assertListEqual([], result['names'])

    async def test_request_attestation(self):
        """
        Check that requesting an attestation works.
        """
        b64_subject_key = (await self.make_request(self.node(0), 'identity/my_peer0/public_key', 'get'))['public_key']
        result = await self.make_request(self.node(1), 'identity/my_peer1/public_key', 'get')
        qb64_authority_key = urllib.parse.quote(result['public_key'], safe='')

        await self.introduce_pseudonyms()

        request = await self.make_request(self.node(0), f'identity/my_peer0/request/{qb64_authority_key}', 'put',
                                          json={
                                              "name": "My attribute",
                                              "schema": "id_metadata",
                                              "metadata": {}})
        self.assertTrue(request['success'])

        outstanding = await self.wait_for(self.node(1), 'identity/my_peer1/outstanding/attestations', 'get')

        self.assertEqual(1, len(outstanding['requests']))
        self.assertEqual(b64_subject_key, outstanding['requests'][0]['peer'])
        self.assertEqual("My attribute", outstanding['requests'][0]['attribute_name'])
        self.assertDictEqual({}, json.loads(outstanding['requests'][0]['metadata']))

    async def test_request_attestation_metadata(self):
        """
        Check that requesting an attestation with metadata works.
        """
        b64_subject_key = (await self.make_request(self.node(0), 'identity/my_peer0/public_key', 'get'))['public_key']
        result = await self.make_request(self.node(1), 'identity/my_peer1/public_key', 'get')
        qb64_authority_key = urllib.parse.quote(result['public_key'], safe='')

        await self.introduce_pseudonyms()

        request = await self.make_request(self.node(0), f'identity/my_peer0/request/{qb64_authority_key}', 'put',
                                          json={
                                              "name": "My attribute",
                                              "schema": "id_metadata",
                                              "metadata": {"Some key": "Some value"}})
        self.assertTrue(request['success'])

        outstanding = await self.wait_for(self.node(1), 'identity/my_peer1/outstanding/attestations', 'get')

        self.assertEqual(1, len(outstanding['requests']))
        self.assertEqual(b64_subject_key, outstanding['requests'][0]['peer'])
        self.assertEqual("My attribute", outstanding['requests'][0]['attribute_name'])
        self.assertDictEqual({"Some key": "Some value"}, json.loads(outstanding['requests'][0]['metadata']))

    async def test_attest(self):
        """
        Check that attesting to an attestation request with metadata works.
        """
        b64_subject_key = (await self.make_request(self.node(0), 'identity/my_peer0/public_key', 'get'))['public_key']
        qb64_subject_key = urllib.parse.quote(b64_subject_key, safe='')
        b64_authority_key = (await self.make_request(self.node(1), 'identity/my_peer1/public_key', 'get'))['public_key']
        qb64_authority_key = urllib.parse.quote(b64_authority_key, safe='')
        metadata = {"Some key": "Some value"}
        request = {"name": "My attribute", "schema": "id_metadata", "metadata": metadata}

        await self.introduce_pseudonyms()
        await self.make_request(self.node(0), f'identity/my_peer0/request/{qb64_authority_key}', 'put', json=request)
        await self.wait_for(self.node(1), 'identity/my_peer1/outstanding/attestations', 'get')

        result = await self.make_request(self.node(1), f'identity/my_peer1/attest/{qb64_subject_key}', 'put',
                                         json={
                                             "name": "My attribute",
                                             "value": base64.b64encode(b'Some value').decode()})
        self.assertTrue(result['success'])

        # How node 0 sees itself after receiving the attestation
        result = await self.wait_for(self.node(0), 'identity/my_peer0/credentials', 'get')
        self.assertEqual(1, len(result['names']))
        self.assertEqual("My attribute", result['names'][0]['name'])
        self.assertListEqual([b64_authority_key], result['names'][0]['attesters'])
        for k, v in metadata.items():
            self.assertIn(k, result['names'][0]['metadata'])
            self.assertEqual(v, result['names'][0]['metadata'][k])

        # How node 1 sees node 0 after making the attestation
        result = await self.wait_for(self.node(1), f'identity/my_peer1/credentials/{qb64_subject_key}', 'get')
        self.assertEqual(1, len(result['names']))
        self.assertEqual("My attribute", result['names'][0]['name'])
        self.assertListEqual([b64_authority_key], result['names'][0]['attesters'])
        for k, v in metadata.items():
            self.assertIn(k, result['names'][0]['metadata'])
            self.assertEqual(v, result['names'][0]['metadata'][k])

    async def test_verify(self):
        """
        Check that verifying a credential works.
        """
        self.nodes.append(await self.create_node())  # We need a third node for this one

        b64_subject_key = (await self.make_request(self.node(0), 'identity/my_peer0/public_key', 'get'))['public_key']
        qb64_subject_key = urllib.parse.quote(b64_subject_key, safe='')
        b64_authority_key = (await self.make_request(self.node(1), 'identity/my_peer1/public_key', 'get'))['public_key']
        qb64_authority_key = urllib.parse.quote(b64_authority_key, safe='')
        b64_verifier_key = (await self.make_request(self.node(2), 'identity/my_peer2/public_key', 'get'))['public_key']
        qb64_verifier_key = urllib.parse.quote(b64_verifier_key, safe='')

        metadata = {"Some key": "Some value"}
        request = {"name": "My attribute", "schema": "id_metadata", "metadata": metadata}
        attest = {"name": "My attribute", "value": base64.b64encode(b'Some value').decode()}

        await self.introduce_pseudonyms()
        await self.make_request(self.node(0), f'identity/my_peer0/request/{qb64_authority_key}', 'put', json=request)
        await self.wait_for(self.node(1), 'identity/my_peer1/outstanding/attestations', 'get')
        await self.make_request(self.node(1), f'identity/my_peer1/attest/{qb64_subject_key}', 'put', json=attest)

        credentials = await self.wait_for(self.node(0), 'identity/my_peer0/credentials', 'get')
        attribute_hash = credentials["names"][0]["hash"]

        result = await self.make_request(self.node(2), f'identity/my_peer2/verify/{qb64_subject_key}', 'put',
                                         json={
                                             "hash": attribute_hash,
                                             "value": attest["value"],
                                             "schema": request["schema"]})
        self.assertTrue(result['success'])

        verification_requests = await self.wait_for_requests(self.node(0),
                                                             f'identity/my_peer0/outstanding/verifications',
                                                             'get')
        self.assertEqual(b64_verifier_key, verification_requests[0]['peer'])
        self.assertEqual("My attribute", verification_requests[0]['attribute_name'])

        result = await self.make_request(self.node(0), f'identity/my_peer0/allow/{qb64_verifier_key}', 'put',
                                         json={"name": "My attribute"})
        self.assertTrue(result['success'])

        await self.deliver_messages()
        output = await self.make_request(self.node(2), 'identity/my_peer2/verifications', 'get')

        self.assertEqual(1, len(output['outputs']))
        self.assertEqual(attribute_hash, output['outputs'][0]['hash'])
        self.assertEqual(base64.b64encode(b'Some value').decode(), output['outputs'][0]['reference'])
        self.assertGreaterEqual(output['outputs'][0]['match'], 0.98)

    async def test_disallow_verify(self):
        """
        Check that no verification is performed, if not allowed.
        """
        self.nodes.append(await self.create_node())  # We need a third node for this one

        b64_subject_key = (await self.make_request(self.node(0), 'identity/my_peer0/public_key', 'get'))['public_key']
        qb64_subject_key = urllib.parse.quote(b64_subject_key, safe='')
        b64_authority_key = (await self.make_request(self.node(1), 'identity/my_peer1/public_key', 'get'))['public_key']
        qb64_authority_key = urllib.parse.quote(b64_authority_key, safe='')
        b64_verifier_key = (await self.make_request(self.node(2), 'identity/my_peer2/public_key', 'get'))['public_key']
        qb64_verifier_key = urllib.parse.quote(b64_verifier_key, safe='')

        metadata = {"Some key": "Some value"}
        request = {"name": "My attribute", "schema": "id_metadata", "metadata": metadata}
        attest = {"name": "My attribute", "value": base64.b64encode(b'Some value').decode()}

        await self.introduce_pseudonyms()
        await self.make_request(self.node(0), f'identity/my_peer0/request/{qb64_authority_key}', 'put', json=request)
        await self.wait_for(self.node(1), 'identity/my_peer1/outstanding/attestations', 'get')
        await self.make_request(self.node(1), f'identity/my_peer1/attest/{qb64_subject_key}', 'put', json=attest)

        credentials = await self.wait_for(self.node(0), 'identity/my_peer0/credentials', 'get')
        attribute_hash = credentials["names"][0]["hash"]

        result = await self.make_request(self.node(2), f'identity/my_peer2/verify/{qb64_subject_key}', 'put',
                                         json={
                                             "hash": attribute_hash,
                                             "value": attest["value"],
                                             "schema": request["schema"]})
        self.assertTrue(result['success'])

        verification_requests = await self.wait_for_requests(self.node(0),
                                                             f'identity/my_peer0/outstanding/verifications',
                                                             'get')
        self.assertEqual(b64_verifier_key, verification_requests[0]['peer'])
        self.assertEqual("My attribute", verification_requests[0]['attribute_name'])

        result = await self.make_request(self.node(0), f'identity/my_peer0/disallow/{qb64_verifier_key}', 'put',
                                         json={"name": "My attribute"})
        self.assertTrue(result['success'])

        await self.deliver_messages()
        output = await self.make_request(self.node(2), 'identity/my_peer2/verifications', 'get')

        self.assertEqual(0, len(output['outputs']))



from ..base import TestBase


class TestDHTBase(TestBase):

    def routing_table(self, i):
        return self.nodes[i].overlay.get_routing_table(self.my_peer(i))

    def storage(self, i):
        return self.nodes[i].overlay.get_storage(self.my_peer(i))

    def my_node_id(self, i):
        return self.nodes[i].overlay.get_my_node_id(self.my_peer(i))



from ..REST.rest_base import RESTTestBase
from ..mocking.community import MockCommunity
from ..mocking.endpoint import MockEndpoint, MockEndpointListener
from ...bootstrapping.dispersy.bootstrapper import DispersyBootstrapper
from ...configuration import DISPERSY_BOOTSTRAPPER
from ...messaging.anonymization.community import TunnelCommunity


class MockSettings:

    def __init__(self):
        self.peer_flags = {0}


class MockTunnelCommunity(TunnelCommunity, MockCommunity):  # pylint: disable=R0901

    def __init__(self):  # pylint: disable=W0231
        # We don't actually initialize the TunnelCommunity, we just want it as a base class (pylint doesn't approve).
        MockCommunity.__init__(self)
        self.settings = MockSettings()
        self.circuits = {}
        self.relay_from_to = {}
        self.exit_sockets = {}

        bootstrapper = DispersyBootstrapper(DISPERSY_BOOTSTRAPPER['init']['ip_addresses'], [])
        self.bootstrappers = [bootstrapper]


class TestOverlaysEndpoint(RESTTestBase):

    FAKE_BOOTSTRAP_ADDRESS = ("127.0.0.1", 0)

    async def setUp(self):  # pylint: disable=W0236
        super().setUp()

        self.fake_endpoint = MockEndpoint(("0.0.0.0", 0), TestOverlaysEndpoint.FAKE_BOOTSTRAP_ADDRESS)
        self.fake_endpoint.open()
        self.fake_endpoint_listener = MockEndpointListener(self.fake_endpoint)
        self.fake_endpoint.add_listener(self.fake_endpoint_listener)

        await self.initialize([], 1)
        self.ipv8 = self.node(0)
        self.ipv8.overlay = MockTunnelCommunity()
        self.ipv8.overlay.network = self.ipv8.network
        self.ipv8.overlays.append(self.ipv8.overlay)

    def bootstrap_addresses(self):
        return self.ipv8.overlay.bootstrappers[0].ip_addresses

    async def test_no_ip(self):
        """
        Test if requests that do not specify an IP are rejected.
        """
        response = await self.make_request(self.ipv8, "isolation", "POST", json={"port": 5, "exitnode": 1},
                                           expected_status=400)

        self.assertFalse(response["success"])

    async def test_no_port(self):
        """
        Test if requests that do not specify a port are rejected.
        """
        response = await self.make_request(self.ipv8, "isolation", "POST", json={"ip": "127.0.0.1", "exitnode": 1},
                                           expected_status=400)

        self.assertFalse(response["success"])

    async def test_no_choice(self):
        """
        Test if requests that do not specify a to add either an exit node or a bootstrap server.
        """
        response = await self.make_request(self.ipv8, "isolation", "POST", json={"ip": "127.0.0.1", "port": 5},
                                           expected_status=400)

        self.assertFalse(response["success"])

    async def test_add_bootstrap(self):
        """
        Check if bootstrap nodes are correctly added.

        A successfully added bootstrap node is walked to.
        """
        ip, port = TestOverlaysEndpoint.FAKE_BOOTSTRAP_ADDRESS

        response = await self.make_request(self.ipv8, "isolation", "POST",
                                           json={"ip": ip, "port": port, "bootstrapnode": 1})

        self.assertTrue(response["success"])
        self.assertIn(TestOverlaysEndpoint.FAKE_BOOTSTRAP_ADDRESS, self.bootstrap_addresses())
        self.assertIn(TestOverlaysEndpoint.FAKE_BOOTSTRAP_ADDRESS, self.ipv8.network.blacklist)
        self.assertLessEqual(1, len(self.fake_endpoint_listener.received_packets))

    async def test_add_exit(self):
        """
        Check if exit nodes are correctly added.

        A successfully added exit node is walked to.
        """
        ip, port = TestOverlaysEndpoint.FAKE_BOOTSTRAP_ADDRESS

        response = await self.make_request(self.ipv8, "isolation", "POST",
                                           json={"ip": ip, "port": port, "exitnode": 1})

        self.assertTrue(response["success"])
        self.assertLessEqual(1, len(self.fake_endpoint_listener.received_packets))



from __future__ import annotations

import binascii
import hashlib
import struct
import typing

from ..signed_object import AbstractSignedObject
from ...types import PrivateKey


class Token(AbstractSignedObject):
    """
    A double pointer, pointing to a previous token and content.
    The contents belonging to the content pointer can be added to a Token object.

    Tokens do not and should not contain an index (but occupy one in a chain).
    Tokens do not and should not contain a reference to the public key (but are signed by one).
    """

    def __init__(self,
                 previous_token_hash: bytes,
                 content: typing.Optional[bytes] = None,
                 content_hash: typing.Optional[bytes] = None,
                 private_key: typing.Optional[PrivateKey] = None,
                 signature: typing.Optional[bytes] = None) -> None:
        """
        Create a new Token object, always specify the hash of the preceding Token.

        For a Token belonging to yourself, give the content and private key and the content hash and signature will
        be automatically created.
        For a Token of someone else or your own reloaded Token, give the content hash and the signature.

        :param previous_token_hash: a string pointing to the hash of the preceding Token.
        :param content: the content to tokenize.
        :param content_hash: the hash of the tokenized content.
        :param private_key: the private key of the current user, creating this new Token.
        :param signature: the signature for this Token.
        """
        if content is not None and content_hash is None:
            self.content: typing.Optional[bytes] = content
            self.content_hash = hashlib.sha3_256(content).digest()
        elif content is None and content_hash is not None:
            self.content = None
            self.content_hash = content_hash
        else:
            raise RuntimeError("Specify either content or content_hash!")

        self.previous_token_hash = previous_token_hash

        super(Token, self).__init__(private_key, signature)

    def get_plaintext(self) -> bytes:
        """
        Return the double pointer of this Token in plain text format, being the concatenation of two hashes.
        """
        return self.previous_token_hash + self.content_hash

    @classmethod
    def unserialize(cls, data, public_key, offset=0) -> Token:
        sig_len = public_key.get_signature_length()
        previous_token_hash, content_hash, signature = struct.unpack_from(f">32s32s{sig_len}s", data, offset=offset)
        return Token(previous_token_hash, content_hash=content_hash, signature=signature)

    def receive_content(self, content: bytes) -> bool:
        """
        Attempt to receive some content, only add it if its hash matches what we expect.

        :param content: the content potentially belonging to the hash pointed to by this Token.
        :returns: whether we accepted the content.
        """
        content_hash = hashlib.sha3_256(content).digest()
        if content_hash == self.content_hash:
            self.content = content
            return True
        return False

    @classmethod
    def create(cls, previous_token: Token, content: bytes, private_key: PrivateKey) -> Token:
        return Token(previous_token.get_hash(), content, private_key=private_key)

    def to_database_tuple(self) -> typing.Tuple[bytes, bytes, bytes, typing.Optional[bytes]]:
        """
        Get a representation of this Token as four byte strings (previous hash, signature, content hash and content).

        :returns: the four byte strings for database insertion.
        """
        return self.previous_token_hash, self.signature, self.content_hash, self.content

    @classmethod
    def from_database_tuple(cls,
                            previous_token_hash: bytes,
                            signature: bytes,
                            content_hash: bytes,
                            content: typing.Optional[bytes]) -> Token:
        """
        Create a Token from a four-byte-string representation (previous hash, signature, content hash and content).

        :param previous_token_hash: the hash of the preceding Token.
        :param signature: the signature over the plaintext Token.
        :param content_hash: the content hash of this Token.
        :param content: optionally the content belonging to this Token.
        """
        token = Token(previous_token_hash, content_hash=content_hash, signature=signature)
        if content is not None:
            token.receive_content(content)
        return token

    def __str__(self) -> str:
        return (f"Token[{binascii.hexlify(self.get_hash()).decode()}]"
                f"({binascii.hexlify(self.previous_token_hash).decode()}, "
                f"{binascii.hexlify(self.content_hash).decode()})")



import logging
from asyncio import ensure_future
from binascii import hexlify, unhexlify

from aiohttp import web

from aiohttp_apispec import docs

from .base_endpoint import BaseEndpoint, HTTP_NOT_FOUND, Response
from .schema import DefaultResponseSchema
from ..dht import DHTError
from ..dht.community import DHTCommunity


class NoBlockDHTEndpoint(BaseEndpoint):
    """
    This endpoint is responsible for handling requests for DHT data, non-blocking.
    """

    def __init__(self):
        super(NoBlockDHTEndpoint, self).__init__()
        self.dht = None

    def setup_routes(self):
        self.app.add_routes([web.get('/{mid}', self.handle_get)])

    def initialize(self, session):
        super(NoBlockDHTEndpoint, self).initialize(session)
        self.dht = session.get_overlay(DHTCommunity)

    @docs(
        tags=["DHT"],
        summary="Connect to a peer through the DHT.",
        parameters=[{
            'in': 'path',
            'name': 'mid',
            'description': 'The mid (i.e., sha1(public_key)) of the peer to connect to.',
            'type': 'string',
            'required': True
        }],
        responses={
            200: {
                "schema": DefaultResponseSchema,
                "examples": {'Success': {"success": True}}
            },
            HTTP_NOT_FOUND: {
                "schema": DefaultResponseSchema,
                "examples": {'DHT not loaded': {"success": False, "error": "DHT community not found"}}
            }
        }
    )
    async def handle_get(self, request):
        if not self.dht:
            return Response({"error": "DHT community not found"}, status=HTTP_NOT_FOUND)

        mid = unhexlify(request.match_info['mid'])

        async def connect_peer():
            try:
                self.dht.connect_peer(mid)
            except DHTError:
                logging.error("DHT Failed to connect to %s", hexlify(mid))
            else:
                logging.error("DHT connected to %s", hexlify(mid))

        ensure_future(connect_peer())
        return Response({"success": True})



"""
This module provides basic database functionalty and simple version control.

@author: Boudewijn Schoon
@organization: Technical University Delft
@contact: dispersy@frayja.com
"""
import logging
import os
import sys
import typing
from abc import ABCMeta, abstractmethod
from collections import defaultdict
from threading import RLock

if sys.platform == "darwin":
    # Workaround for annoying MacOS Sierra bug: https://bugs.python.org/issue27126
    # As fix, we are using pysqlite2 so we can supply our own version of sqlite3.
    try:
        import pysqlite2.dbapi2 as sqlite3
    except ImportError:
        # For newer installations, we can just import it
        from sqlite3 import dbapi2 as sqlite3
else:
    import sqlite3


def execute_or_script(cursor, statement):
    """
    This workaround is part of the MacOS Sierra bug described at the top of this file.
    """
    if sys.platform == "darwin":
        cursor.executescript(statement)
    else:
        cursor.execute(statement)


# In Python 3 sqlite expects bytes instead of buffer objects.
database_blob = bytes
db_locks: typing.Dict[str, RLock] = defaultdict(RLock)


def db_call(f):
    def wrapper(self, *args, **kwargs):
        with db_locks[self._file_path]:
            if self._cursor:
                return f(self, *args, **kwargs)
            return None
    return wrapper


def _thread_safe_result_it(result, fetch_all=True):
    rows = (result.fetchall() if fetch_all else result.fetchone()) or []
    return (row for row in rows)


class IgnoreCommits(Exception):

    """
    Ignore all commits made within the body of a 'with database:' clause.

    with database:
       # all commit statements are delayed until the database.__exit__
       database.commit()
       database.commit()
       # raising IgnoreCommits causes all commits to be ignored
       raise IgnoreCommits()
    """
    def __init__(self):
        super(IgnoreCommits, self).__init__("Ignore all commits made within __enter__ and __exit__")


class DatabaseException(RuntimeError):
    pass


class Database(metaclass=ABCMeta):

    def __init__(self, file_path):
        """
        Initialize a new Database instance.

        @param file_path: the path to the database file.
        @type file_path: unicode
        """
        self._assert(isinstance(file_path, str),
                     "expected file_path to be unicode, but was %s" % str(type(file_path)))

        super(Database, self).__init__()
        self._logger = logging.getLogger(self.__class__.__name__)

        self._logger.debug("loading database [%s]", file_path)
        self._file_path = file_path

        # _CONNECTION, _CURSOR, AND _DATABASE_VERSION are set during open(...)
        self._connection = None
        self._cursor = None
        self._database_version = 0

        # Database.commit() is enabled when _pending_commits == 0.  Database.commit() is disabled
        # when _pending_commits > 0.  A commit is required when _pending_commits > 1.
        self._pending_commits = 0

    def _assert(self, condition, message=""):
        """
        Check if condition is True, or raise a DatabaseException with a message.
        """
        if not condition:
            raise DatabaseException(str(message))

    def open(self, initial_statements=True, prepare_visioning=True):
        self._assert(self._cursor is None, "Database.open() has already been called")
        self._assert(self._connection is None, "Database.open() has already been called")

        self._logger.debug("open database [%s]", self._file_path)
        if (not self._file_path.startswith(':')) and (not os.path.isfile(self._file_path)):
            if not os.path.exists(os.path.dirname(self._file_path)):
                os.makedirs(os.path.dirname(self._file_path))
        self._connect()
        if initial_statements:
            self._initial_statements()
        if prepare_visioning:
            self._prepare_version()
        return True

    @db_call
    def close(self, commit=True):
        self._assert(self._cursor is not None,
                     "Database.close() has been called or Database.open() has not been called")
        self._assert(self._connection is not None,
                     "Database.close() has been called or Database.open() has not been called")
        if commit:
            self.commit(exiting=True)
        self._logger.debug("close database [%s]", self._file_path)
        self._cursor.close()
        self._cursor = None
        self._connection.close()
        self._connection = None
        return True

    def _connect(self):
        self._connection = sqlite3.connect(self._file_path, check_same_thread=False)
        self._connection.text_factory = bytes
        self._cursor = self._connection.cursor()

        assert self._cursor

    def _initial_statements(self):
        self._assert(self._cursor is not None,
                     "Database.close() has been called or Database.open() has not been called")
        self._assert(self._connection is not None,
                     "Database.close() has been called or Database.open() has not been called")

        # collect current database configuration
        page_size = int(next(self._cursor.execute("PRAGMA page_size"))[0])
        journal_mode = next(self._cursor.execute("PRAGMA journal_mode"))[0].decode().upper()
        synchronous = next(self._cursor.execute("PRAGMA synchronous"))[0]
        synchronous = synchronous.decode().upper() if isinstance(synchronous, bytes) else synchronous

        #
        # PRAGMA page_size = bytes;
        # http://www.sqlite.org/pragma.html#pragma_page_size
        # Note that changing page_size has no effect unless performed on a new database or followed
        # directly by VACUUM.  Since we do not want the cost of VACUUM every time we load a
        # database, existing databases must be upgraded.
        #
        if page_size < 8192:
            self._logger.debug("PRAGMA page_size = 8192 (previously: %s) [%s]", page_size, self._file_path)

            # it is not possible to change page_size when WAL is enabled
            if journal_mode == "WAL":
                self._cursor.executescript("PRAGMA journal_mode = DELETE")
                journal_mode = "DELETE"
            self._cursor.execute("PRAGMA page_size = 8192")
            execute_or_script(self._cursor, "VACUUM")
            page_size = 8192

        else:
            self._logger.debug("PRAGMA page_size = %s (no change) [%s]", page_size, self._file_path)

        #
        # PRAGMA journal_mode = DELETE | TRUNCATE | PERSIST | MEMORY | WAL | OFF
        # http://www.sqlite.org/pragma.html#pragma_page_size
        #
        if not (journal_mode == "WAL" or self._file_path == ":memory:"):
            self._logger.debug("PRAGMA journal_mode = WAL (previously: %s) [%s]", journal_mode, self._file_path)
            self._cursor.execute("PRAGMA locking_mode = EXCLUSIVE")
            execute_or_script(self._cursor, "PRAGMA journal_mode = WAL")

        else:
            self._logger.debug("PRAGMA journal_mode = %s (no change) [%s]", journal_mode, self._file_path)

        #
        # PRAGMA synchronous = 0 | OFF | 1 | NORMAL | 2 | FULL;
        # http://www.sqlite.org/pragma.html#pragma_synchronous
        #
        if synchronous not in ("NORMAL", 1):
            self._logger.debug("PRAGMA synchronous = NORMAL (previously: %s) [%s]", synchronous, self._file_path)
            execute_or_script(self._cursor, "PRAGMA synchronous = NORMAL")

        else:
            self._logger.debug("PRAGMA synchronous = %s (no change) [%s]", synchronous, self._file_path)

    def _prepare_version(self):
        self._assert(self._cursor is not None,
                     "Database.close() has been called or Database.open() has not been called")
        self._assert(self._connection is not None,
                     "Database.close() has been called or Database.open() has not been called")

        # check is the database contains an 'option' table
        try:
            count, = next(self.execute("SELECT COUNT(*) FROM sqlite_master WHERE type = 'table' AND name = 'option'"))
        except StopIteration:
            raise RuntimeError()

        if count:
            # get version from required 'option' table
            try:
                version, = next(self.execute("SELECT value FROM option WHERE key == 'database_version' LIMIT 1"))
            except StopIteration:
                # the 'database_version' key was not found
                version = b"0"
        else:
            # the 'option' table probably hasn't been created yet
            version = b"0"

        self._database_version = self.check_database(version)
        self._assert(isinstance(self._database_version, int),
                     "expected databse version to be int or long, but was type %s" % str(type(self._database_version)))

    @property
    def database_version(self):
        return self._database_version

    @property
    def file_path(self):
        """
        The database filename including path.
        """
        return self._file_path

    def __enter__(self):
        """
        Enters a no-commit state.  The commit will be performed by __exit__.

        @return: The method self.execute
        """
        self._assert(self._cursor is not None,
                     "Database.close() has been called or Database.open() has not been called")
        self._assert(self._connection is not None,
                     "Database.close() has been called or Database.open() has not been called")

        self._logger.debug("disabling commit [%s]", self._file_path)
        self._pending_commits = max(1, self._pending_commits)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Leaves a no-commit state.  A commit will be performed if Database.commit() was called while
        in the no-commit state.
        """
        self._assert(self._cursor is not None,
                     "Database.close() has been called or Database.open() has not been called")
        self._assert(self._connection is not None,
                     "Database.close() has been called or Database.open() has not been called")

        self._pending_commits, pending_commits = 0, self._pending_commits

        if exc_type is None:
            self._logger.debug("enabling commit [%s]", self._file_path)
            if pending_commits > 1:
                self._logger.debug("performing %d pending commits [%s]", pending_commits - 1, self._file_path)
                self.commit()
            return True

        elif isinstance(exc_value, IgnoreCommits):
            self._logger.debug("enabling commit without committing now [%s]", self._file_path)
            return True

        else:
            # Niels 23-01-2013, an exception happened from within the with database block
            # returning False to let Python reraise the exception.
            return False

    @db_call
    def execute(self, statement, bindings=(), get_lastrowid=False, fetch_all=True):
        """
        Execute one SQL statement.

        A SQL query must be presented in unicode format.  This is to ensure that no unicode
        exeptions occur when the bindings are merged into the statement.

        Furthermore, the bindings may not contain any strings either.  For a 'string' the unicode
        type must be used.  For a binary string the buffer(...) type must be used.

        The SQL query may contain placeholder entries defined with a '?'.  Each of these
        placeholders will be used to store one value from bindings.  The placeholders are filled by
        sqlite and all proper escaping is done, making this the preferred way of adding variables to
        the SQL query.

        @param statement: the SQL statement that is to be executed.
        @type statement: unicode

        @param bindings: the values that must be set to the placeholders in statement.
        @type bindings: list, tuple, dict, or set

        @returns: unknown
        @raise sqlite.Error: unknown
        """
        self._logger.log(logging.NOTSET, "%s <-- %s [%s]", statement, bindings, self._file_path)
        result = self._cursor.execute(statement, bindings)
        if get_lastrowid:
            return self._cursor.lastrowid
        return _thread_safe_result_it(result, fetch_all)

    @db_call
    def executescript(self, statements, fetch_all=True):
        self._assert(self._cursor is not None,
                     "Database.close() has been called or Database.open() has not been called")
        self._assert(self._connection is not None,
                     "Database.close() has been called or Database.open() has not been called")
        self._assert(isinstance(statements, str), "The SQL statement must be given in unicode")

        self._logger.log(logging.NOTSET, "%s [%s]", statements, self._file_path)

        result = self._cursor.executescript(statements)
        return _thread_safe_result_it(result, fetch_all)

    @db_call
    def executemany(self, statement, sequenceofbindings, fetch_all=True):
        """
        Execute one SQL statement several times.

        All SQL queries must be presented in unicode format.  This is to ensure that no unicode
        exeptions occur when the bindings are merged into the statement.

        Furthermore, the bindings may not contain any strings either.  For a 'string' the unicode
        type must be used.  For a binary string the buffer(...) type must be used.

        The SQL query may contain placeholder entries defined with a '?'.  Each of these
        placeholders will be used to store one value from bindings.  The placeholders are filled by
        sqlite and all proper escaping is done, making this the preferred way of adding variables to
        the SQL query.

        @param statement: the SQL statement that is to be executed.
        @type statement: unicode

        @param sequenceofbindings: a list, tuple, set, or generator of bindings, where every binding
                                   contains the values that must be set to the placeholders in
                                   statement.

        @type sequenceofbindings: list, tuple, set or generator

        @returns: unknown
        @raise sqlite.Error: unknown
        """
        self._assert(self._cursor is not None,
                     "Database.close() has been called or Database.open() has not been called")
        self._assert(self._connection is not None,
                     "Database.close() has been called or Database.open() has not been called")

        self._logger.log(logging.NOTSET, "%s [%s]", statement, self._file_path)
        result = self._cursor.executemany(statement, sequenceofbindings)
        return _thread_safe_result_it(result, fetch_all)

    @db_call
    def commit(self, exiting=False):
        self._assert(self._cursor is not None,
                     "Database.close() has been called or Database.open() has not been called")
        self._assert(self._connection is not None,
                     "Database.close() has been called or Database.open() has not been called")
        self._assert(not (exiting and self._pending_commits), "No pending commits should be present when exiting")

        if self._pending_commits:
            self._logger.debug("defer commit [%s]", self._file_path)
            self._pending_commits += 1
            return False

        else:
            self._logger.debug("commit [%s]", self._file_path)
            return self._connection.commit()

    @abstractmethod
    def check_database(self, database_version):
        """
        Check the database and upgrade if required.

        This method is called once for each Database instance to ensure that the database structure
        and version is correct.  Each Database must contain one table of the structure below where
        the database_version is stored.  This value is used to keep track of the current database
        version.

        >>> CREATE TABLE option(key TEXT PRIMARY KEY, value BLOB);
        >>> INSERT INTO option(key, value) VALUES('database_version', '1');

        @param database_version: the current database_version value from the option table. This
         value reverts to u'0' when the table could not be accessed.
        @type database_version: unicode
        """
        pass



from binascii import hexlify

from aiohttp import web

from aiohttp_apispec import docs, json_schema

from marshmallow.fields import Boolean, Dict, List, Nested, String

from .base_endpoint import BaseEndpoint, HTTP_BAD_REQUEST, HTTP_PRECONDITION_FAILED, Response
from .schema import DefaultResponseSchema, OverlaySchema, OverlayStatisticsSchema, schema
from ..messaging.interfaces.statistics_endpoint import StatisticsEndpoint


class OverlaysEndpoint(BaseEndpoint):
    """
    This endpoint is responsible for handing all requests regarding the status of overlays.
    """

    def __init__(self):
        super(OverlaysEndpoint, self).__init__()
        self.statistics_supported = None

    def setup_routes(self):
        self.app.add_routes([web.get('', self.get_overlays),
                             web.get('/statistics', self.get_statistics),
                             web.post('/statistics', self.enable_statistics)])

    def initialize(self, session):
        super(OverlaysEndpoint, self).initialize(session)
        self.statistics_supported = isinstance(session.endpoint, StatisticsEndpoint) \
            or isinstance(getattr(session.endpoint, 'endpoint', None), StatisticsEndpoint)

    @docs(
        tags=["Overlays"],
        summary="Return information about all currently loaded overlays.",
        responses={
            200: {
                "schema": schema(OverlayResponse={
                    "overlays": [OverlaySchema]
                })
            }
        }
    )
    async def get_overlays(self, _):
        overlay_stats = []
        for overlay in self.session.overlays:
            peers = overlay.get_peers()
            statistics = self.session.endpoint.get_aggregate_statistics(overlay.get_prefix()) \
                if isinstance(self.session.endpoint, StatisticsEndpoint) else {}
            overlay_stats.append({
                "id": hexlify(overlay.community_id).decode('utf-8'),
                "my_peer": hexlify(overlay.my_peer.public_key.key_to_bin()).decode('utf-8'),
                "global_time": overlay.global_time,
                "peers": [{'ip': peer.address[0],
                           'port': peer.address[1],
                           'public_key': hexlify(peer.public_key.key_to_bin()).decode('utf-8')} for peer in peers],
                "overlay_name": overlay.__class__.__name__,
                "statistics": statistics,
                "max_peers": overlay.max_peers,
                "is_isolated": self.session.network != overlay.network,
                "my_estimated_wan": {"ip": overlay.my_estimated_wan[0], "port": overlay.my_estimated_wan[1]},
                "my_estimated_lan": {"ip": overlay.my_estimated_lan[0], "port": overlay.my_estimated_lan[1]},
                "strategies": [{'name': strategy.__class__.__name__,
                                'target_peers': target_peers}
                               for strategy, target_peers in self.session.strategies if strategy.overlay == overlay]
            })
        return Response({"overlays": overlay_stats})

    @docs(
        tags=["Overlays"],
        summary="Return statistics for all currently loaded overlays.",
        responses={
            200: {
                "schema": schema(StatisticsResponse={
                    "statistics": List(Dict(keys=String, values=Nested(OverlayStatisticsSchema))),
                }),
                "examples": {'Success': {"statistics": [{"DiscoveryCommunity": {'num_up': 0, 'num_down': 0,
                                                                                'bytes_up': 0, 'bytes_down': 0,
                                                                                'diff_time': 0}}]}}
            }
        }
    )
    async def get_statistics(self, _):
        overlay_stats = []
        for overlay in self.session.overlays:
            statistics = self.session.endpoint.get_statistics(overlay.get_prefix()) if self.statistics_supported else {}
            overlay_stats.append({
                overlay.__class__.__name__: self.statistics_by_name(statistics, overlay)
            })
        return Response({"statistics": overlay_stats})

    def statistics_by_name(self, statistics, overlay):
        named_statistics = {}
        for message_id, network_stats in statistics.items():
            if overlay.decode_map[message_id]:
                mapped_name = str(message_id) + ":" + overlay.decode_map[message_id].__name__
            else:
                mapped_name = str(message_id) + ":unknown"
            mapped_value = network_stats.to_dict()
            named_statistics[mapped_name] = mapped_value
        return named_statistics

    @docs(
        tags=["Overlays"],
        summary="Enable/disable statistics for a given overlay.",
        responses={
            200: {
                "schema": DefaultResponseSchema,
                "examples": {'Success': {"success": True}}
            },
            HTTP_PRECONDITION_FAILED: {
                "schema": DefaultResponseSchema,
                "examples": {'Statistics disabled': {"success": False, "error": "StatisticsEndpoint is not enabled."}}
            },
            HTTP_BAD_REQUEST: {
                "schema": DefaultResponseSchema,
                "examples": {'Missing parameter': {"success": False, "error": "Parameter 'enable' is required."}}
            }
        }
    )
    @json_schema(schema(EnableStatisticsRequest={
        'enable*': (Boolean, 'Whether to enable or disable the statistics'),
        'all': (Boolean, 'Whether update applies to all overlays'),
        'overlay_name': (String, 'Class name of the overlay'),
    }))
    async def enable_statistics(self, request):
        if not self.statistics_supported:
            return Response({"success": False, "error": "StatisticsEndpoint is not enabled."},
                            status=HTTP_PRECONDITION_FAILED)

        args = await request.json()
        if 'enable' not in args:
            return Response({"success": False, "error": "Parameter 'enable' is required"}, status=HTTP_BAD_REQUEST)
        if 'all' not in args and 'overlay_name' not in args:
            return Response({"success": False, "error": "Parameter 'all' or 'overlay_name' is required"},
                            status=HTTP_PRECONDITION_FAILED)

        self.enable_overlay_statistics(enable=args['enable'],
                                       class_name=args.get('overlay_name', None),
                                       all_overlays=args.get('all', False))
        return Response({"success": True})

    def enable_overlay_statistics(self, enable=False, class_name=None, all_overlays=False):
        if all_overlays:
            for overlay in self.session.overlays:
                self.session.endpoint.enable_community_statistics(overlay.get_prefix(), enable)
        elif class_name:
            for overlay in self.session.overlays:
                if overlay.__class__.__name__ == class_name:
                    self.session.endpoint.enable_community_statistics(overlay.get_prefix(), enable)



from . import DHTError

# Sentinel object
Null = object()


class Node(object):
    """
    This class represents a node within a prefix tree.
    """

    def __init__(self):
        self.value = None
        self.children = {}


class Trie(object):
    """
    This class represents a prefix tree.
    """

    def __init__(self, alphabet):
        self.alphabet = alphabet
        self.root = Node()

    def _find(self, key):
        node = self.root
        for char in key:
            node = node.children.get(char)
            if node is None:
                break
        return node

    def __getitem__(self, key):
        node = self._find(key)
        if node is None or node.value is None:
            raise KeyError
        return node.value

    def __setitem__(self, key, value):
        node = self.root
        for char in key:
            if char not in self.alphabet:
                raise DHTError('Error while adding item to trie')

            next_node = node.children.get(char)
            if next_node is None:
                next_node = node.children[char] = Node()
            node = next_node
        node.value = value

    def __delitem__(self, key):
        toremove = []

        node = self.root
        toremove.append(('', node))
        for char in key:
            toremove.append((char, node))
            node = node.children.get(char)
            if node is None:
                break

        if node is None or node.value is None:
            raise KeyError

        node.value = None
        while node.value is None and not node.children and toremove:
            char, node = toremove.pop()
            node.children.pop(char)

    def itervalues(self):
        def generator(node):
            if node.value is not None:
                yield node.value
            for _, child in node.children.items():
                for subresult in generator(child):
                    yield subresult
        return generator(self.root)

    def values(self):
        return list(self.itervalues())

    def longest_prefix_item(self, key, default=Null):
        prefix = ''
        value = None

        node = self.root
        for index, _ in enumerate(key):
            node = node.children.get(key[index])
            if node is None:
                break
            if node.value is not None:
                prefix = key[:index + 1]
                value = node.value

        if value:
            return prefix, value
        elif default is not Null:
            return default
        raise KeyError

    def longest_prefix(self, key, default=Null):
        result = self.longest_prefix_item(key, default=default)
        return result[0] if result != default else default

    def longest_prefix_value(self, key, default=Null):
        result = self.longest_prefix_item(key, default=default)
        return result[1] if result != default else default

    def suffixes(self, key):
        node = self._find(key)

        suffixes = []
        if node is None:
            return suffixes
        if node.value is not None:
            suffixes.append('')

        for char, node in node.children.items():
            if node.value:
                suffixes.append(char)
            for nested_suffix in self.suffixes(key + char):
                suffix = char + nested_suffix
                if suffix not in suffixes:
                    suffixes.append(suffix)

        return suffixes



from .base import TestBase
from ..database import Database


class MockDatabase(Database):

    def check_database(self, database_version):
        self.execute("CREATE TABLE option(key TEXT PRIMARY KEY, value BLOB)")
        self.execute("INSERT INTO option(key, value) VALUES('database_version', '0')")
        self.commit()
        return 0


class TestDatabase(TestBase):

    def setUp(self):
        super(TestDatabase, self).setUp()
        self.database = MockDatabase(u":memory:")

    def test_unloaded(self):
        """
        Check if an unloaded database returns None for queries.
        """
        self.assertIsNone(self.database.execute("SELECT * FROM option"))

    def test_closed(self):
        """
        Check if an unloaded database returns None for queries.
        """
        self.database.open()
        self.assertListEqual([(b'database_version', b'0')], list(self.database.execute("SELECT * FROM option")))
        self.database.close(True)
        self.assertIsNone(self.database.execute("SELECT * FROM option"))



from .routing import Node
from ..messaging.lazy_payload import VariablePayload, vp_compile


@vp_compile
class PingRequestPayload(VariablePayload):
    msg_id = 1
    names = ['identifier']
    format_list = ['I']


@vp_compile
class PingResponsePayload(VariablePayload):
    msg_id = 2
    names = ['identifier']
    format_list = ['I']


@vp_compile
class StoreRequestPayload(VariablePayload):
    msg_id = 3
    names = ['identifier', 'token', 'target', 'values']
    format_list = ['I', '20s', '20s', 'varlenH-list']


@vp_compile
class StoreResponsePayload(VariablePayload):
    msg_id = 4
    names = ['identifier']
    format_list = ['I']


@vp_compile
class FindRequestPayload(VariablePayload):
    msg_id = 5
    names = ['identifier', 'lan_address', 'target', 'offset', 'force_nodes']
    format_list = ['I', 'ip_address', '20s', 'I', '?']


@vp_compile
class FindResponsePayload(VariablePayload):
    msg_id = 6
    names = ['identifier', 'token', 'values', 'nodes']
    format_list = ['I', '20s', 'varlenH-list', 'node-list']


@vp_compile
class StorePeerRequestPayload(VariablePayload):
    msg_id = 7
    names = ['identifier', 'token', 'target']
    format_list = ['I', '20s', '20s']


@vp_compile
class StorePeerResponsePayload(VariablePayload):
    msg_id = 8
    names = ['identifier']
    format_list = ['I']


@vp_compile
class ConnectPeerRequestPayload(VariablePayload):
    msg_id = 9
    names = ['identifier', 'lan_address', 'target']
    format_list = ['I', 'ip_address', '20s']


@vp_compile
class ConnectPeerResponsePayload(VariablePayload):
    msg_id = 10
    names = ['identifier', 'nodes']
    format_list = ['I', 'node-list']


@vp_compile
class StrPayload(VariablePayload):
    names = ['data']
    format_list = ['raw']


@vp_compile
class SignedStrPayload(VariablePayload):
    names = ['data', 'version', 'public_key']
    format_list = ['varlenH', 'I', 'varlenH']


class NodePacker:
    def __init__(self, serializer):
        self.serializer = serializer

    def pack(self, node):
        return self.serializer.pack('ip_address', node.address) + \
            self.serializer.pack('varlenH', node.public_key.key_to_bin())

    def unpack(self, data, offset, unpack_list):
        address, offset = self.serializer.unpack('ip_address', data, offset)
        key, offset = self.serializer.unpack('varlenH', data, offset)
        unpack_list.append(Node(key, address=address))
        return offset



from aiohttp import web

from aiohttp_apispec import docs, json_schema

from marshmallow.fields import Boolean, Integer, String

from .base_endpoint import BaseEndpoint, HTTP_BAD_REQUEST, Response
from .schema import DefaultResponseSchema, schema
from ..bootstrapping.dispersy.bootstrapper import DispersyBootstrapper
from ..messaging.anonymization.community import TunnelCommunity


class IsolationEndpoint(BaseEndpoint):
    """
    This endpoint is responsible for on-demand adding of addresses for different services.
    """

    def setup_routes(self):
        self.app.add_routes([web.post('', self.handle_post)])

    def add_exit_node(self, address):
        for overlay in self.session.overlays:
            if isinstance(overlay, TunnelCommunity):
                overlay.walk_to(address)

    def add_bootstrap_server(self, address):
        if hasattr(self.session, "network"):
            self.session.network.blacklist.append(address)
        for overlay in self.session.overlays:
            overlay.network.blacklist.append(address)
            overlay.walk_to(address)
            for bootstrapper in overlay.bootstrappers:
                if isinstance(bootstrapper, DispersyBootstrapper):
                    bootstrapper.ip_addresses.append(address)

    @docs(
        tags=["Isolation"],
        summary="Add an address to a specific IPv8 service.",
        responses={
            200: {
                "schema": DefaultResponseSchema,
                "examples": {'Success': {"success": True}}
            },
            HTTP_BAD_REQUEST: {
                "schema": DefaultResponseSchema,
                "examples": {'Bad IPv4 address': {"success": False, "error": "Traceback (most recent call last): ..."}}
            }
        }
    )
    @json_schema(schema(IsolationRequest={
        'ip*': String,
        'port*': Integer,
        'bootstrapnode': Boolean,
        'exitnode': Boolean
    }))
    async def handle_post(self, request):
        # Check if we have arguments, containing an address and the type of address to add.
        args = await request.json()
        if not args or 'ip' not in args or 'port' not in args:
            return Response({"success": False, "error": "Parameters 'ip' and 'port' are required"},
                            status=HTTP_BAD_REQUEST)
        if 'exitnode' not in args and 'bootstrapnode' not in args:
            return Response({"success": False, "error": "Parameter 'exitnode' or 'bootstrapnode' is required"},
                            status=HTTP_BAD_REQUEST)
        address_str = args['ip']
        port_num = args['port']
        fmt_address = (address_str, port_num)
        # Actually add the address to the requested service
        if 'exitnode' in args:
            self.add_exit_node(fmt_address)
        else:
            self.add_bootstrap_server(fmt_address)
        return Response({"success": True})



from base64 import b64encode
from binascii import hexlify, unhexlify
from timeit import default_timer

from aiohttp import web

from aiohttp_apispec import docs, json_schema

from marshmallow.fields import Integer, String

from .base_endpoint import BaseEndpoint, HTTP_BAD_REQUEST, HTTP_NOT_FOUND, Response
from .schema import DHTValueSchema, DefaultResponseSchema, schema
from ..dht import DHTError
from ..dht.community import DHTCommunity
from ..dht.discovery import DHTDiscoveryCommunity
from ..dht.routing import calc_node_id
from ..messaging.interfaces.dispatcher.endpoint import FAST_ADDR_TO_INTERFACE


class DHTEndpoint(BaseEndpoint):
    """
    This endpoint is responsible for handling requests for DHT data.
    """

    def __init__(self):
        super(DHTEndpoint, self).__init__()
        self.dht = self.publisher = None

    def setup_routes(self):
        self.app.add_routes([web.get('/statistics', self.get_statistics),
                             web.get('/values', self.get_stored_values),
                             web.get('/values/{key}', self.get_values),
                             web.put('/values/{key}', self.put_value),
                             web.get('/peers/{mid}', self.get_peer),
                             web.get('/buckets', self.get_buckets),
                             web.get('/buckets/{prefix:\\w*}/refresh', self.refresh_bucket)])

    def initialize(self, session):
        super(DHTEndpoint, self).initialize(session)
        self.dht = session.get_overlay(DHTCommunity)

    @docs(
        tags=["DHT"],
        summary="Return DHT statistics.",
        responses={
            200: {
                "schema": schema(DHTStatsResponse={
                    "statistics": schema(DHTStats={
                        "node_id": String,
                        "peer_id": String,
                        "routing_table_size": Integer,
                        "routing_table_buckets": Integer,
                        "num_keys_in_store": Integer,
                        "num_tokens": Integer,
                        "num_peers_in_store": Integer,
                        "num_store_for_me": Integer
                    })
                })
            },
            400: {
                "schema": DefaultResponseSchema,
                "examples": {'DHT disabled': {"success": False, "error": "DHT community not found."}}
            }
        }
    )
    async def get_statistics(self, _):
        if not self.dht:
            return Response({"success": False, "error": "DHT community not found"}, status=HTTP_NOT_FOUND)

        stats = {
            "peer_id": hexlify(self.dht.my_peer.mid).decode('utf-8'),
            "num_tokens": len(self.dht.tokens),
            "endpoints": []
        }

        for address_cls, routing_table in self.dht.routing_tables.items():
            buckets = routing_table.trie.values()
            address = self.dht.my_peer.addresses.get(address_cls, self.dht.my_estimated_wan)
            stats["endpoints"].append({
                "endpoint": FAST_ADDR_TO_INTERFACE[address_cls],
                "node_id": hexlify(calc_node_id(address, self.dht.my_peer.mid)).decode('utf-8'),
                "routing_table_size": sum([len(bucket.nodes) for bucket in buckets]),
                "routing_table_buckets": len(buckets),
                "num_keys_in_store":
                    len(self.dht.storages.get(address_cls).items) if self.dht.storages.get(address_cls) else 0,
            })

        if isinstance(self.dht, DHTDiscoveryCommunity):
            stats.update({
                "num_peers_in_store": {hexlify(key).decode('utf-8'): len(peers)
                                       for key, peers in self.dht.store.items()},
                "num_store_for_me": {hexlify(key).decode('utf-8'): len(peers)
                                     for key, peers in self.dht.store_for_me.items()}
            })

        return Response({"statistics": stats})

    @docs(
        tags=["DHT"],
        summary="Connect to a peer using the DHT.",
        responses={
            200: {
                "schema": schema(DHTPeerResponse={
                    "peers": [schema(DHTPeer={
                        "public_key": String,
                        "address": String
                    })]
                })
            },
            400: {
                "schema": DefaultResponseSchema,
                "examples": {'DHT disabled': {"success": False, "error": "DHT community not found."}}
            }
        }
    )
    async def get_peer(self, request):
        if not self.dht:
            return Response({"success": False, "error": "DHT community not found"}, status=HTTP_NOT_FOUND)

        mid = unhexlify(request.match_info['mid'])
        nodes = await self.dht.connect_peer(mid)
        return Response({"peers": [{'public_key': b64encode(node.public_key.key_to_bin()).decode('utf-8'),
                                    'address': node.address} for node in nodes]})

    @docs(
        tags=["DHT"],
        summary="Get a list of locally stored key-value pairs from the DHT.",
        responses={
            200: {
                "schema": schema(DHTStoredValuesResponse={
                    "values": [DHTValueSchema]
                })
            },
            400: {
                "schema": DefaultResponseSchema,
                "examples": {'DHT disabled': {"success": False, "error": "DHT community not found."}}
            }
        }
    )
    async def get_stored_values(self, _):
        if not self.dht:
            return Response({"success": False, "error": "DHT community not found"}, status=HTTP_NOT_FOUND)

        results = {}
        for address_cls, storage in self.dht.storages.items():
            for key, raw_values in storage.items.items():
                values = self.dht.post_process_values([v.data for v in raw_values])
                dicts = []
                for value in values:
                    data, public_key = value
                    dicts.append({
                        'endpoint': FAST_ADDR_TO_INTERFACE[address_cls],
                        'public_key': b64encode(public_key).decode('utf-8') if public_key else None,
                        'key': hexlify(key).decode('utf-8'),
                        'value': hexlify(data).decode('utf-8')
                    })
                results[hexlify(key).decode()] = dicts
        return Response(results)

    @docs(
        tags=["DHT"],
        summary="Lookup the values for a specific key on the DHT.",
        responses={
            200: {
                "schema": schema(DHTValuesResponse={
                    "values": [DHTValueSchema]
                })
            },
            400: {
                "schema": DefaultResponseSchema,
                "examples": {'DHT disabled': {"success": False, "error": "DHT community not found."}}
            }
        }
    )
    async def get_values(self, request):
        if not self.dht:
            return Response({"success": False, "error": "DHT community not found"}, status=HTTP_NOT_FOUND)

        key = unhexlify(request.match_info['key'])

        start = default_timer()
        values, crawls = await self.dht.find_values(key, debug=True)
        nodes_tried = set().union(*[crawl.nodes_tried for crawl in crawls])
        responses = sum([crawl.responses for crawl in crawls], [])
        stop = default_timer()

        return Response({
            "values": [{'public_key': b64encode(public_key).decode('utf-8') if public_key else None,
                        'key': hexlify(key).decode('utf-8'),
                        'value': hexlify(data).decode('utf-8')} for data, public_key in values],
            "debug": {
                "requests": len(nodes_tried),
                "responses": len(responses),
                "responses_with_nodes": len([r for s, r in responses if 'nodes' in r]),
                "responses_with_values": len([r for s, r in responses if 'values' in r]),
                "time": stop - start
            }
        })

    @docs(
        tags=["DHT"],
        summary="Store a key-value pair on the DHT.",
        parameters=[{
            'in': 'path',
            'name': 'key',
            'description': 'The key under which to store the value',
            'type': 'string',
            'required': True
        }],
        responses={
            200: {"schema": DefaultResponseSchema},
            400: {
                "schema": DefaultResponseSchema,
                "examples": {'DHT disabled': {"success": False, "error": "DHT community not found."}}
            }
        }
    )
    @json_schema(schema(DHTStoreRequest={
        'value*': String
    }))
    async def put_value(self, request):
        if not self.dht:
            return Response({"success": False, "error": "DHT community not found"}, status=HTTP_NOT_FOUND)

        parameters = await request.json()
        if 'value' not in parameters:
            return Response({"success": False, "error": "incorrect parameters"}, status=HTTP_BAD_REQUEST)

        key = unhexlify(request.match_info['key'])
        await self.dht.store_value(key, unhexlify(parameters['value']), sign=True)
        return Response({"success": True})

    @docs(
        tags=["DHT"],
        summary="Return a list of all buckets in the routing table of the DHT community.",
        responses={
            200: {
                "schema": schema(BucketsResponse={
                    "buckets": [schema(Bucket={
                        "prefix": String,
                        "last_changed": Integer,
                        "peers": [schema(BucketPeer={
                            "ip": String,
                            "port": Integer,
                            "mid": String,
                            "id": String,
                            "failed": Integer,
                            "last_contact": Integer,
                            "distance": Integer
                        })]
                    })]
                })
            }
        }
    )
    async def get_buckets(self, _):
        return Response({"buckets": [{
            "prefix": bucket.prefix_id,
            "last_changed": bucket.last_changed,
            "endpoint": FAST_ADDR_TO_INTERFACE[address_cls],
            "peers": [{
                "ip": peer.address[0],
                "port": peer.address[1],
                "mid": hexlify(peer.mid).decode('utf-8'),
                "id": hexlify(peer.id).decode('utf-8'),
                "failed": peer.failed,
                "last_contact": peer.last_contact,
                "distance": peer.distance(self.dht.get_my_node_id(peer)),
            } for peer in bucket.nodes.values()]
        } for address_cls, routing_table in self.dht.routing_tables.items() for bucket in routing_table.trie.values()]})

    @docs(
        tags=["DHT"],
        summary="Refresh a specific bucket in the DHT community.",
        parameters=[{
            'in': 'path',
            'name': 'prefix',
            'description': 'Prefix of the bucket which to refresh.',
            'type': 'string',
            'required': True
        }],
        responses={
            200: {
                "schema": DefaultResponseSchema,
                "example": {"success": True}
            },
            400: {
                "schema": DefaultResponseSchema,
                "examples": {'Unknown bucket': {"success": False, "error": "no such bucket"}}
            }
        }
    )
    async def refresh_bucket(self, request):
        prefix = request.match_info['prefix']
        success = False
        error = None
        for routing_table in list(self.dht.routing_tables.values()):
            try:
                await self.dht.find_values(routing_table.trie[prefix].generate_id())
                success = True
            except KeyError:
                pass
            except DHTError as e:
                error = e

        if not error and not success:
            return Response({"success": False, "error": "no such bucket"}, status=HTTP_BAD_REQUEST)
        if error and not success:
            return Response({"success": False, "error": str(error)})
        return Response({"success": True})



from asyncio import Future, ensure_future, get_event_loop, sleep
from contextlib import suppress

from .base import TestBase
from ..taskmanager import TaskManager, task
from ..util import coroutine


class TestTaskManager(TestBase):

    def setUp(self):
        super(TestTaskManager, self).setUp()
        self.tm = TaskManager()
        self.counter = 0

    async def tearDown(self):
        await self.tm.shutdown_task_manager()
        return await super(TestTaskManager, self).tearDown()

    def test_call_later(self):
        self.tm.register_task("test", lambda: None, delay=10)
        self.assertTrue(self.tm.is_pending_task_active("test"))

    def test_call_later_and_cancel(self):
        self.tm.register_task("test", lambda: None, delay=10)
        self.tm.cancel_pending_task("test")
        self.assertFalse(self.tm.is_pending_task_active("test"))

    async def test_replace_with_duplicate(self):
        task1 = self.tm.register_task("test", lambda: None, delay=10)
        await self.tm.replace_task("test", lambda: None, delay=10)
        await sleep(.1)
        self.assertTrue(self.tm.is_pending_task_active("test"))
        self.assertTrue(task1.cancelled())

    async def test_replace_without_duplicate(self):
        task = await self.tm.replace_task("test", lambda: None, delay=10)
        await sleep(.1)
        self.assertTrue(self.tm.is_pending_task_active("test"))
        self.assertFalse(task.done())

    async def test_replace_with_error(self):
        with self.assertRaises(ValueError):
            await self.tm.replace_task("test", "bad argument", delay=10)

    def test_looping_call(self):
        self.tm.register_task("test", lambda: None, interval=10)
        self.assertTrue(self.tm.is_pending_task_active("test"))

    def test_looping_call_and_cancel(self):
        self.tm.register_task("test", lambda: None, interval=10)
        self.tm.cancel_pending_task("test")
        self.assertFalse(self.tm.is_pending_task_active("test"))

    def test_delayed_looping_call_requires_coroutine(self):
        with self.assertRaises(TypeError):
            self.tm.register_task("test", ensure_future(coroutine(lambda: None)), delay=1)

    def test_delayed_looping_call_register_and_cancel_pre_delay(self):
        self.assertFalse(self.tm.is_pending_task_active("test"))
        self.tm.register_task("test", lambda: None, delay=1, interval=1)
        self.assertTrue(self.tm.is_pending_task_active("test"))
        self.tm.cancel_pending_task("test")
        self.assertFalse(self.tm.is_pending_task_active("test"))

    async def test_delayed_looping_call_register_wait_and_cancel(self):
        self.assertFalse(self.tm.is_pending_task_active("test"))
        self.tm.register_task("test", self.count, interval=.1)
        self.assertTrue(self.tm.is_pending_task_active("test"))
        await sleep(.15)
        self.assertEqual(1, self.counter)
        self.assertTrue(self.tm.is_pending_task_active("test"))
        await sleep(.1)
        self.assertEqual(2, self.counter)
        # After canceling the task the counter should stop increasing
        self.tm.cancel_pending_task("test")
        self.assertFalse(self.tm.is_pending_task_active("test"))
        await sleep(.1)
        self.assertEqual(2, self.counter)

    def test_raise_on_duplicate_task_name(self):
        self.tm.register_task("test", lambda: None)
        with self.assertRaises(RuntimeError):
            self.tm.register_task("test", lambda: None)

    def test_duplicate_anon_task(self):
        task = lambda: None
        self.tm.register_anonymous_task("test", task)
        self.tm.register_anonymous_task("test", task)
        self.assertEqual(2, len(self.tm.get_tasks()))
        self.tm.cancel_all_pending_tasks()

    async def test_shutdown(self):
        """
        Test if the TaskManager does not allow new tasks after shutdown().
        """
        await self.tm.shutdown_task_manager()
        self.tm.register_anonymous_task("test", lambda: None)
        self.assertFalse(self.tm.is_pending_task_active('test'))

    async def test_cleanup(self):
        """
        Test if the tasks are cleaned up after the cleanup frequency has been met.
        """
        await self.tm.register_anonymous_task("test", lambda: None)
        self.assertEqual(0, len(self.tm.get_tasks()))

    async def test_cleanup_remaining(self):
        """
        Test if tasks which have yet to complete are not cleaned.
        """
        await self.tm.register_anonymous_task("test", lambda: None)
        self.tm.register_anonymous_task("test", sleep, 10)
        self.assertEqual(1, len(self.tm.get_tasks()))

    async def test_task_with_exception(self):
        def exception_handler(_, __):
            exception_handler.called = True
        exception_handler.called = False

        get_event_loop().set_exception_handler(exception_handler)
        with suppress(ZeroDivisionError):
            await self.tm.register_task('test', lambda: 1 / 0)
        self.assertTrue(exception_handler.called)

    async def test_task_with_exception_ignore(self):
        def exception_handler(_, __):
            exception_handler.called = True
        exception_handler.called = False

        get_event_loop().set_exception_handler(exception_handler)
        with suppress(ZeroDivisionError):
            await self.tm.register_task('test', lambda: 1 / 0, ignore=(ZeroDivisionError,))
        self.assertFalse(exception_handler.called)

    async def test_task_decorator_coro(self):
        future = Future()

        @task
        async def task_func(_):
            future.set_result(None)
        task_func(self.tm)
        await future

    def test_task_decorator_no_coro(self):
        with self.assertRaises(TypeError):
            @task
            def task_func(_):
                pass

    def count(self):
        self.counter += 1

    def set_counter(self, value):
        self.counter = value



import logging
import time
import traceback
from asyncio import CancelledError, Future, Task, ensure_future, gather, iscoroutinefunction, sleep
from contextlib import suppress
from functools import wraps
from threading import RLock
from weakref import WeakValueDictionary

from .util import coroutine, succeed

MAX_TASK_AGE = 600


async def interval_runner(delay, interval, task, *args):
    await sleep(delay)
    while True:
        await task(*args)
        await sleep(interval)


async def delay_runner(delay, task, *args):
    await sleep(delay)
    await task(*args)


def task(func):
    """
    Register a TaskManager function as an anonymous task and return the Task
    object so that it can be awaited if needed. Any exceptions will be logged.
    Note that if awaited, exceptions will still need to be handled.
    """
    if not iscoroutinefunction(func):
        raise TypeError('Task decorator should be used with coroutine functions only!')

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        return self.register_anonymous_task(func.__name__,
                                            ensure_future(func(self, *args, **kwargs)),
                                            ignore=(Exception,))
    return wrapper


class TaskManager(object):
    """
    Provides a set of tools to maintain a list of asyncio Tasks that are to be
    executed during the lifetime of an arbitrary object, usually getting killed with it.
    """

    def __init__(self):
        self._pending_tasks = WeakValueDictionary()
        self._task_lock = RLock()
        self._shutdown = False
        self._counter = 0
        self._logger = logging.getLogger(self.__class__.__name__)

        self._checker = self.register_task('_check_tasks', self._check_tasks,
                                           interval=MAX_TASK_AGE, delay=MAX_TASK_AGE * 1.5)

    def _check_tasks(self):
        now = time.time()
        for name, task in self._pending_tasks.items():
            if not task.interval and now - task.start_time > MAX_TASK_AGE:
                self._logger.warning('Non-interval task "%s" has been running for %.2f!', name, now - task.start_time)

    def replace_task(self, name, *args, **kwargs):
        """
        Replace named task with the new one, cancelling the old one in the process.
        """
        new_task = Future()

        def cancel_cb(_):
            try:
                new_task.set_result(self.register_task(name, *args, **kwargs))
            except Exception as e:
                new_task.set_exception(e)

        old_task = self.cancel_pending_task(name)
        old_task.add_done_callback(cancel_cb)
        return new_task

    def register_task(self, name, task, *args, delay=None, interval=None, ignore=()):
        """
        Register a Task/(coroutine)function so it can be canceled at shutdown time or by name.
        """
        if not isinstance(task, Task) and not iscoroutinefunction(task) and not callable(task):
            raise ValueError('Register_task takes a Task or a (coroutine)function as a parameter')
        if (interval or delay) and isinstance(task, Task):
            raise ValueError('Cannot run Task at an interval or with a delay')
        if not isinstance(ignore, tuple) or not all((issubclass(e, Exception) for e in ignore)):
            raise ValueError('Ignore should be a tuple of Exceptions or None')

        with self._task_lock:
            if self._shutdown:
                self._logger.warning("Not adding task %s due to shutdown!", str(task))
                if isinstance(task, (Task, Future)):
                    if not task.done():
                        task.cancel()
                # We need to return an awaitable in case the caller awaits the output of register_task.
                return succeed(None)

            if self.is_pending_task_active(name):
                raise RuntimeError("Task already exists: '%s'" % name)

            if iscoroutinefunction(task) or callable(task):
                task = task if iscoroutinefunction(task) else coroutine(task)
                if interval:
                    # The default delay for looping calls is the same as the interval
                    delay = interval if delay is None else delay
                    task = ensure_future(interval_runner(delay, interval, task, *args))
                elif delay:
                    task = ensure_future(delay_runner(delay, task, *args))
                else:
                    task = ensure_future(task(*args))
            # Since weak references to list/tuple are not allowed, we're not storing start_time/interval
            # in _pending_tasks. Instead we add them as attributes to the task.
            task.start_time = time.time()
            task.interval = interval
            # The set_name function is only available in Python 3.8+
            task_name = f'{self.__class__.__name__}:{name}'
            if hasattr(task, 'set_name'):
                task.set_name(task_name)
            else:
                task.name = task_name

            assert isinstance(task, Task)

            def done_cb(future):
                self._pending_tasks.pop(name, None)
                try:
                    future.result()
                except CancelledError:
                    pass
                except ignore as e:
                    self._logger.error('Task resulted in error: %s\n%s', e, ''.join(traceback.format_exc()))

            self._pending_tasks[name] = task
            task.add_done_callback(done_cb)
            return task

    def register_anonymous_task(self, basename, task, *args, **kwargs):
        """
        Wrapper for register_task to derive a unique name from the basename.
        """
        self._counter += 1
        return self.register_task(basename + ' ' + str(self._counter), task, *args, **kwargs)

    def cancel_pending_task(self, name):
        """
        Cancels the named task
        """
        with self._task_lock:
            task = self._pending_tasks.get(name, None)
            if not task:
                return succeed(None)

            if not task.done():
                task.cancel()
                self._pending_tasks.pop(name, None)
            return task

    def cancel_all_pending_tasks(self):
        """
        Cancels all the registered tasks.
        This usually should be called when stopping or destroying the object so no tasks are left floating around.
        """
        with self._task_lock:
            assert all([isinstance(t, (Task, Future)) for t in self._pending_tasks.values()]), self._pending_tasks
            return [self.cancel_pending_task(name) for name in list(self._pending_tasks.keys())]

    def is_pending_task_active(self, name):
        """
        Return a boolean determining if a task is active.
        """
        with self._task_lock:
            task = self._pending_tasks.get(name, None)
            return not task.done() if task else False

    def get_tasks(self):
        """
        Returns a list of all registered tasks, excluding tasks the are created by the TaskManager itself.
        """
        with self._task_lock:
            return [t for t in self._pending_tasks.values() if t != self._checker]

    async def wait_for_tasks(self):
        """
        Waits until all registered tasks are done.
        """
        tasks = self.get_tasks()
        if tasks:
            await gather(*tasks, return_exceptions=True)

    async def shutdown_task_manager(self):
        """
        Clear the task manager, cancel all pending tasks and disallow new tasks being added.
        """
        with self._task_lock:
            self._shutdown = True
            tasks = self.cancel_all_pending_tasks()

        if tasks:
            with suppress(CancelledError):
                await gather(*tasks)


__all__ = ["TaskManager", "task"]



from .wallet.irmaexact.keydump import nijmegen_pk_1568208470


FORMATS = {
    "id_metadata": {
        "algorithm": "bonehexact",
        "key_size": 32,  # Pairings over 1024 bit space
        "hash": "sha256_4"  # 4 byte hash
    },
    "id_metadata_big": {
        "algorithm": "bonehexact",
        "key_size": 64,  # Pairings over 4096 bit space
        "hash": "sha256"  # 32 byte hash
    },
    "id_metadata_huge": {
        "algorithm": "bonehexact",
        "key_size": 96,  # Pairings over 9216 bit space
        "hash": "sha512"  # 64 byte hash
    },
    "id_metadata_range_18plus": {
        "algorithm": "pengbaorange",
        "key_size": 32,  # Pairings over 1024 bit space
        "min": 18,
        "max": 200
    },
    "id_irma_nijmegen_address_1568208470": {
        "algorithm": "irmaexact",
        "issuer_pk": nijmegen_pk_1568208470,  # Valid until Wednesday 11 September 2019 13:27:50 GMT
        "order": [u"street", u"houseNumber", u"zipcode", u"municipality", u"city"],
        "credential": u"pbdf.nijmegen.address",
        "keyCounter": 0,
        "validity": 13

    },
    "id_irma_nijmegen_personalData_1568208470": {
        "algorithm": "irmaexact",
        "issuer_pk": nijmegen_pk_1568208470,  # Valid until Wednesday 11 September 2019 13:27:50 GMT
        "order": [u"initials", u"firstnames", u"prefix", u"familyname", u"surname",
                  u"fullname", u"dateofbirth", u"gender", u"nationality"],
        "credential": u"pbdf.nijmegen.personalData",
        "keyCounter": 0,
        "validity": 13
    },
    "id_irma_nijmegen_ageLimits_1568208470": {
        "algorithm": "irmaexact",
        "issuer_pk": nijmegen_pk_1568208470,  # Valid until Wednesday 11 September 2019 13:27:50 GMT
        "order": [u"over12", u"over16", u"over18", u"over21", u"over65"],
        "credential": u"pbdf.nijmegen.ageLimits",
        "keyCounter": 0,
        "validity": 13
    },
    "id_irma_nijmegen_bsn_1568208470": {
        "algorithm": "irmaexact",
        "issuer_pk": nijmegen_pk_1568208470,  # Valid until Wednesday 11 September 2019 13:27:50 GMT
        "order": [u"bsn"],
        "credential": u"pbdf.nijmegen.bsn",
        "keyCounter": 0,
        "validity": 13
    }
}



from time import time

from .routing import Node
from ..peer import Peer
from ..peerdiscovery.discovery import DiscoveryStrategy


class PingChurn(DiscoveryStrategy):

    def __init__(self, overlay, ping_interval=25):
        super(PingChurn, self).__init__(overlay)
        self.ping_interval = ping_interval

    def take_step(self):
        with self.walk_lock:
            # Nothing is happening yet, skip this step
            if not self.overlay.routing_tables:
                return

            for routing_table in self.overlay.routing_tables.values():
                for node in routing_table.remove_bad_nodes():
                    self.overlay.network.remove_peer(node)

            for peer in self.overlay.get_peers():
                if peer.address in self.overlay.network.blacklist:
                    continue

                node = Node(peer.key, peer.address)
                routing_table = self.overlay.get_routing_table(node)
                if not routing_table.has(node.id) and not routing_table.add(node):
                    self.overlay.network.remove_peer(peer)

            for routing_table in self.overlay.routing_tables.values():
                for bucket in routing_table.trie.values():
                    for node in bucket.nodes.values():
                        if node not in self.overlay.get_peers():
                            peer = Peer(node.key, node.address)
                            self.overlay.network.add_verified_peer(peer)
                            self.overlay.network.discover_services(peer, [self.overlay.community_id])

            now = time()
            for routing_table in self.overlay.routing_tables.values():
                for bucket in routing_table.trie.values():
                    for node in bucket.nodes.values():
                        if node.last_ping_sent + self.ping_interval <= now:
                            self.overlay.ping(node)



import asyncio
import heapq


class DiscreteLoop(asyncio.AbstractEventLoop):
    """
    A discrete asyncio loop that immediately executes incoming tasks without a real-time waiting period.
    This loop can be helpful when quickly running simulation experiments with the IPv8 library. Usage:

    loop = DiscreteLoop()
    set_event_loop(loop)
    """

    def __init__(self):
        self._time = 0
        self._running = False
        self._immediate = []
        self._scheduled = []
        self._exc = None

    def get_debug(self):
        return False

    def time(self):
        return self._time

    def run_forever(self):
        self._running = True
        asyncio._set_running_loop(self)
        while (self._immediate or self._scheduled) and self._running:
            if self._immediate:
                h = self._immediate[0]
                self._immediate = self._immediate[1:]
            else:
                h = heapq.heappop(self._scheduled)
                self._time = h._when
                h._scheduled = False
            if not h._cancelled:
                h._run()
            if self._exc is not None:
                raise self._exc

    def run_until_complete(self, future):
        raise NotImplementedError

    def _timer_handle_cancelled(self, handle):
        pass

    def is_running(self):
        return self._running

    def is_closed(self):
        return not self._running

    def stop(self):
        self._running = False

    def close(self):
        self._running = False

    def call_exception_handler(self, context):
        self._exc = context.get('exception', None)

    def call_soon(self, callback, *args, context=None):
        h = asyncio.Handle(callback, args, self)
        self._immediate.append(h)
        return h

    def call_later(self, delay, callback, *args):
        if delay < 0:
            raise Exception("Can't schedule in the past")
        return self.call_at(self._time + delay, callback, *args)

    def call_at(self, when, callback, *args):
        if when < self._time:
            raise Exception("Can't schedule in the past")
        h = asyncio.TimerHandle(when, callback, args, self)
        heapq.heappush(self._scheduled, h)
        h._scheduled = True
        return h

    def create_task(self, coro):
        async def wrapper():
            try:
                await coro
            except asyncio.CancelledError:
                pass
            except Exception as e:
                self._exc = e

        return asyncio.Task(wrapper(), loop=self)

    def create_future(self):
        return asyncio.Future(loop=self)



"""
Copyright (c) 2016, Maarten Everts
All rights reserved.

This source code has been ported from https://github.com/privacybydesign/gabi
The authors of this file are not -in any way- affiliated with the original authors or organizations.
"""

from .proofs import ProofD, hashCommit
from .. import secure_randint
from ...primitives.attestation import sha256_as_int
from ...primitives.value import FP2Value


class Credential(object):

    def __init__(self, Pk, Attributes, Signature):
        self.Signature = Signature
        self.Pk = Pk
        self.Attributes = Attributes

    def CreateDisclosureProof(self, disclosedAttributes, context, nonce1):
        undisclosedAttributes = getUndisclosedAttributes(disclosedAttributes, len(self.Attributes))

        randSig = self.Signature.Randomize(self.Pk)

        eCommit = secure_randint(self.Pk.Params.LeCommit)
        vCommit = secure_randint(self.Pk.Params.LvCommit)

        aCommits = {}

        for v in undisclosedAttributes:
            aCommits[v] = secure_randint(self.Pk.Params.LmCommit)

        Ae = FP2Value(self.Pk.N, randSig.A).intpow(eCommit).a
        Sv = FP2Value(self.Pk.N, self.Pk.S).intpow(vCommit).a
        Z = (Ae * Sv) % self.Pk.N

        for v in undisclosedAttributes:
            Z = (Z * FP2Value(self.Pk.N, self.Pk.R[v]).intpow(aCommits[v]).a) % self.Pk.N

        c = hashCommit([context, randSig.A, Z, nonce1], False)

        ePrime = randSig.E - (1 << (self.Pk.Params.Le - 1))
        eResponse = c * ePrime + eCommit
        vResponse = c * randSig.V + vCommit

        aResponses = {}
        for v in undisclosedAttributes:
            exp = self.Attributes[v]
            if exp.bit_length() > self.Pk.Params.Lm:
                exp = sha256_as_int(str(exp))
            t = c * exp
            aResponses[v] = t + aCommits[v]

        aDisclosed = {}
        for v in disclosedAttributes:
            aDisclosed[v] = self.Attributes[v]

        return ProofD(c, randSig.A, eResponse, vResponse, aResponses, aDisclosed)

    def CreateDisclosureProofBuilder(self, disclosedAttributes):
        return DisclosureProofBuilder(self.Signature.Randomize(self.Pk),
                                      secure_randint(self.Pk.Params.LeCommit),
                                      secure_randint(self.Pk.Params.LvCommit),
                                      {v: secure_randint(self.Pk.Params.LmCommit) for v
                                       in getUndisclosedAttributes(disclosedAttributes, len(self.Attributes))},
                                      1,
                                      disclosedAttributes,
                                      getUndisclosedAttributes(disclosedAttributes, len(self.Attributes)),
                                      self.Pk,
                                      self.Attributes)


class DisclosureProofBuilder(object):

    def __init__(self, randomizedSignature, eCommit, vCommit, attrRandomizers, z, disclosedAttributes,
                 undisclosedAttributes, pk, attributes):
        self.randomizedSignature = randomizedSignature
        self.eCommit = eCommit
        self.vCommit = vCommit
        self.attrRandomizers = attrRandomizers
        self.z = z
        self.disclosedAttributes = disclosedAttributes
        self.undisclosedAttributes = undisclosedAttributes
        self.pk = pk
        self.attributes = attributes

    def MergeProofPCommitment(self, commitment):
        self.z = (self.z * commitment.Pcommit) % self.pk.N

    def PublicKey(self):
        return self.pk

    def Commit(self, skRandomizer):
        self.attrRandomizers[0] = skRandomizer

        Ae = FP2Value(self.pk.N, self.randomizedSignature.A).intpow(self.eCommit).a
        Sv = FP2Value(self.pk.N, self.pk.S).intpow(self.vCommit).a
        self.z = (self.z * Ae * Sv) % self.pk.N

        for v in self.undisclosedAttributes:
            self.z = (self.z * FP2Value(self.pk.N, self.pk.R[v]).intpow(self.attrRandomizers[v]).a) % self.pk.N

        return [self.randomizedSignature.A, self.z]

    def CreateProof(self, challenge):
        ePrime = self.randomizedSignature.E - (1 << (self.pk.Params.Le - 1))
        eResponse = challenge * ePrime + self.eCommit
        vResponse = challenge * self.randomizedSignature.V + self.vCommit

        aResponses = {}
        for v in self.undisclosedAttributes:
            exp = self.attributes[v]
            if exp.bit_length() > self.pk.Params.Lm:
                exp = sha256_as_int(str(exp))
            aResponses[v] = challenge * exp + self.attrRandomizers[v]

        aDisclosed = {v: self.attributes[v] for v in self.disclosedAttributes}

        return ProofD(challenge, self.randomizedSignature.A, eResponse, vResponse, aResponses, aDisclosed)

    def TimestampRequestContributions(self):
        disclosed = [0] * len(self.attributes)
        for i in self.disclosedAttributes:
            disclosed[i] = self.attributes[i]
        return self.randomizedSignature.A, disclosed


def getUndisclosedAttributes(disclosedAttributes, numAttributes):
    return list(set(range(numAttributes)) - set(disclosedAttributes))



from binascii import hexlify
from math import sqrt
from os import urandom

from .boudot import EL, SQR
from .structs import PengBaoAttestation, PengBaoCommitment, PengBaoCommitmentPrivate, PengBaoPublicData


def _random_number(bytelen):
    return int(hexlify(urandom(bytelen)), 16)


def create_attest_pair(PK, value, a, b, bitspace):  # pylint: disable=R0914
    """
    Create an proof that a <= value <= b, for a public key's value lying within a certain bitspace.
    """
    bytespace = bitspace // 8
    r = _random_number(bytespace)
    ra = _random_number(bytespace)
    raa = _random_number(bitspace // 16)
    raa = raa * raa

    w = _random_number(bytespace)
    w2 = w * w

    c = PK.g.intpow(value) * PK.h.intpow(r)

    c1 = c // (PK.g.intpow(a - 1))
    c2 = PK.g.intpow(b + 1) // c
    ca = c1.intpow(b - value + 1) * PK.h.intpow(ra)
    caa = ca.intpow(w2) * PK.h.intpow(raa)

    mst = w2 * (value - a + 1) * (b - value + 1)
    m4 = 0
    while not m4:
        m4 = _random_number(bytespace) % (int(sqrt(mst)) - 1)
    m3 = m4 * m4
    m1 = 0
    while not m1:
        m1 = _random_number(bytespace) % (mst - m4)
    m2 = mst - m1 - m3

    rst = w2 * ((b - value + 1) * r + ra) + raa
    r1 = 0
    while not r1:
        r1 = _random_number(bytespace * bytespace) % (rst // 2 - 1)
    r2 = 0
    while not r2:
        r2 = _random_number(bytespace * bytespace) % (rst // 2 - 1)
    r3 = rst - r1 - r2

    ca1 = PK.g.intpow(m1) * PK.h.intpow(r1)
    ca2 = PK.g.intpow(m2) * PK.h.intpow(r2)
    ca3 = caa // (ca1 * ca2)

    el = EL.create(b - value + 1, -r, ra, PK.g, PK.h, c1, PK.h, b, bitspace)
    sqr1 = SQR.create(w, raa, ca, PK.h, b, bitspace)
    sqr2 = SQR.create(m4, r3, PK.g, PK.h, b, bitspace)

    publicdata = PengBaoPublicData(PK, bitspace, PengBaoCommitment(c, c1, c2, ca, ca1, ca2, ca3, caa), el, sqr1, sqr2)
    privatedata = PengBaoCommitmentPrivate(m1, m2, m3, r1, r2, r3)

    return PengBaoAttestation(publicdata, privatedata)



import typing

from .attestation import Attestation
from .metadata import Metadata
from ..tokentree.token import Token
from ...database import Database
from ...types import PublicKey


class Credential(object):
    """
    Cache for Metadata <- [Attestation] mappings.
    """

    def __init__(self, metadata: Metadata, attestations: typing.Set[Attestation]):
        self.metadata = metadata
        self.attestations = attestations


def to_list(obj) -> list:
    """
    If no results are found in the database, the cursor will return None.
    This function always returns a list.
    """
    return [] if obj is None else list(obj)


class IdentityDatabase(Database):
    """
    Database to store identity layer data.

    This database stores: meta information of different pseudonyms, both ours and those of others.
    This database does not store: commitments and schemes for proving knowledge of data.
    """

    LATEST_DB_VERSION = 1

    def insert_token(self, public_key: PublicKey, token: Token) -> None:
        """
        Inject a Token belonging to a certain public key into this database.
        """
        previous_token_hash, signature, content_hash, content = token.to_database_tuple()
        self.execute("INSERT OR IGNORE INTO Tokens "
                     "(public_key, previous_token_hash, signature, content_hash, content) "
                     "VALUES(?,?,?,?,?)",
                     (public_key.key_to_bin(), previous_token_hash, signature, content_hash, content))
        self.commit()

    def insert_metadata(self, public_key: PublicKey, metadata: Metadata) -> None:
        """
        Inject Metadata belonging to a certain public key into this database.
        """
        token_pointer, signature, serialized_json_dict = metadata.to_database_tuple()
        self.execute("INSERT OR IGNORE INTO Metadata "
                     "(public_key, token_pointer, signature, serialized_json_dict) "
                     "VALUES(?,?,?,?)",
                     (public_key.key_to_bin(), token_pointer, signature, serialized_json_dict))
        self.commit()

    def insert_attestation(self, public_key: PublicKey, authority_key: PublicKey, attestation: Attestation) -> None:
        """
        Inject an Attestation made by some authority for a certain public key into this database.
        """
        metadata_pointer, signature = attestation.to_database_tuple()
        self.execute("INSERT OR IGNORE INTO Attestations "
                     "(public_key, authority_key, metadata_pointer, signature) "
                     "VALUES(?,?,?,?)",
                     (public_key.key_to_bin(), authority_key.key_to_bin(), metadata_pointer, signature))
        self.commit()

    def get_tokens_for(self, public_key: PublicKey) -> typing.Set[Token]:
        """
        Get all tokens in the tree of a certain public key.
        """
        tokens = to_list(self.execute("SELECT previous_token_hash, signature, content_hash, content "
                                      "FROM Tokens WHERE public_key = ?", (public_key.key_to_bin(),),
                                      fetch_all=True))
        return {Token.from_database_tuple(*token) for token in tokens}

    def get_metadata_for(self, public_key: PublicKey) -> typing.Set[Metadata]:
        """
        Get all known metadata for a certain public key.
        """
        metadata = to_list(self.execute("SELECT token_pointer, signature, serialized_json_dict "
                                        "FROM Metadata WHERE public_key = ?", (public_key.key_to_bin(),),
                                        fetch_all=True))
        return {Metadata.from_database_tuple(*metadato) for metadato in metadata}

    def get_attestations_for(self, public_key: PublicKey) -> typing.Set[Attestation]:
        """
        Get all known attestations (made by others) for a certain public key.
        """
        attestations = to_list(self.execute("SELECT metadata_pointer, signature FROM Attestations WHERE public_key = ?",
                                            (public_key.key_to_bin(),), fetch_all=True))
        return {Attestation.from_database_tuple(*attestation) for attestation in attestations}

    def get_attestations_by(self, public_key: PublicKey) -> typing.Set[Attestation]:
        """
        Get all attestations made by a certain public key (for others).

        This is the signing authority.
        """
        attestations = to_list(self.execute("SELECT metadata_pointer, signature "
                                            "FROM Attestations WHERE authority_key = ?",
                                            (public_key.key_to_bin(),), fetch_all=True))
        return {Attestation.from_database_tuple(*attestation) for attestation in attestations}

    def get_attestations_over(self, metadata: Metadata) -> typing.Set[Attestation]:
        """
        Get all known attestations for given metadata.
        """
        attestations = to_list(self.execute("SELECT metadata_pointer, signature FROM Attestations "
                                            "WHERE metadata_pointer = ?",
                                            (metadata.get_hash(),), fetch_all=True))
        return {Attestation.from_database_tuple(*attestation) for attestation in attestations}

    def get_authority(self, attestation: Attestation) -> bytes:
        """
        Retrieve the authority that created a certain attestation.
        """
        return next(self.execute("SELECT authority_key from Attestations WHERE signature = ?",
                                 (attestation.signature,), fetch_all=False))

    def get_credential_over(self, metadata: Metadata) -> Credential:
        """
        Collect all attestations for the given metadata, forming a credential.
        """
        return Credential(metadata, self.get_attestations_over(metadata))

    def get_credentials_for(self, public_key: PublicKey) -> typing.List[Credential]:
        """
        Get all credentials for a given public key.
        """
        return [Credential(metadata, self.get_attestations_over(metadata))
                for metadata in self.get_metadata_for(public_key)]

    def get_known_identities(self):
        """
        List the public keys of all known identity owners.
        """
        out = []
        for result in self.execute("SELECT public_key FROM Tokens", fetch_all=True):
            # These are single item tuples
            out.append(result[0])
        return out

    def get_schema(self, version: int) -> str:
        """
        Return the schema for the database.
        """
        schema = """
                 CREATE TABLE IF NOT EXISTS Tokens(
                 public_key BLOB,
                 previous_token_hash BLOB,
                 signature BLOB,
                 content_hash BLOB,
                 content LONGBLOB,

                 PRIMARY KEY (public_key, previous_token_hash, content_hash)
                 );

                 CREATE TABLE IF NOT EXISTS Metadata(
                 public_key BLOB,
                 token_pointer BLOB,
                 signature BLOB,
                 serialized_json_dict LONGBLOB,

                 PRIMARY KEY (public_key, token_pointer)
                 );

                 CREATE TABLE IF NOT EXISTS Attestations(
                 public_key BLOB,
                 authority_key BLOB,
                 metadata_pointer BLOB,
                 signature BLOB,

                 PRIMARY KEY (public_key, metadata_pointer)
                 );
                 """
        schema += """
                  CREATE TABLE IF NOT EXISTS option(key TEXT PRIMARY KEY, value BLOB);
                  DELETE FROM option WHERE key = 'database_version';
                  INSERT INTO option(key, value) VALUES('database_version', '%s');
                  """ % str(self.LATEST_DB_VERSION)
        return schema

    def check_database(self, database_version: str) -> int:
        assert database_version.isdigit()
        assert int(database_version) >= 0
        database_version_num = int(database_version) or self.LATEST_DB_VERSION

        # This is where an existing schema would be upgraded.
        # As no changes have been made, there is nothing to upgrade.

        self.executescript(self.get_schema(database_version_num))
        self.commit()

        return self.LATEST_DB_VERSION
